"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Any, Optional, List

import importlib
from functools import total_ordering
from pathlib import Path

from spil.sid.core import query_helper
from spil.util.caching import lru_cache as cache
from spil.util.log import debug, info, warning

from spil import conf
from spil.util.exception import SpilException

"""
TODO:

Notes.
We need a STRICT mode, where type changes are warned, or changes to untyped are errored.
Example:
an erroneous usage of get_with() can silently change the Sid type to an unexpected one, 
or loose the type, and the error goes undetected.
We could use a target type for Sids operations.
We could use a global data validation framework (pydantic, apischema or so).
"""


class BaseSid:
    """Base class for Sids.

    Implements __new__ and __init__ to allow Sid instances to be created by a factory.
    The goal is to have an extendable Sid and factory.
    """

    def __new__(cls, *args, **kwargs):
        """
        Calls a Factory function that handles the Sid object creation.
        Returns the instance from that factory.

        (To avoid that the factory calls itself,
        the method uses the kwarg "from_factory".
        If it is set, an object instance is returned.)

        Args:
            *args:
            **kwargs:
        """
        if not kwargs.get("from_factory"):
            (mod, fn) = cls._factory
            mod = importlib.import_module(mod)
            fn = getattr(mod, fn)
            return fn(*args, **kwargs)
        else:
            return object.__new__(cls)

    def __init__(self, *args, **kwargs):
        """
        Empty Method.
        See documentation about the Sid Factory creation mechanism.

        Args:
            *args:
            **kwargs:
        """
        pass


@total_ordering
class StringSid(BaseSid):
    """
    StringSid is the barest form of Sids.
    It only has a string.
    It is not typed.
    """

    _string = ""

    @property
    def string(self) -> str:
        """
        Returns the string representation of a Sid.
        The string can be any arbitrary string if the Sid is not typed.

        If the Sid is typed, the string will is a path like representation of its values.

        Examples:

            >>> Sid('hamlet/s/sq030/sh0100/anim').string
            'hamlet/s/sq030/sh0100/anim'

            >>> Sid('blablabla').string
            'blablabla'

        Returns:
            The string representation.
        """
        return self._string

    @property
    def uri(self) -> str:
        """
        Returns the "uri" representation of a Sid.
        The form is "type:string" if typed, else just the sid string.

        (This property is overriden in TypedSid)

        Returns:
            The uri representation.
        """
        return self._string

    def copy(self) -> Sid:
        """
        Returns a copy of the current Sid.
        Copy is generated by calling `Sid(self.uri)`.

        Example:

            >>> Sid('hamlet/a/char/ophelia/model').copy()
            Sid('asset__task:hamlet/a/char/ophelia/model')

        Returns:
            A Sid copy.
        """
        return Sid(self.uri)  # self.get_with()

    def is_search(self) -> bool:
        """
        Returns True if the current Sid contains search symbols.
        Search symbols are defined in the Sid conf, and are typically: ['*', ',', '>', '<', '**']

        The Sid may not be typed.

        Examples:

            >>> Sid('hamlet/a/char/*/*').is_search()
            True

            >>> Sid('hamlet/a/char/ophelia').is_search()
            False

        Returns:
            True if the Sid contains search symbols, else False.

        """
        return any(s in str(self) for s in conf.search_symbols)

    def __str__(self) -> str:
        return self._string

    def __repr__(self) -> str:
        return "Sid('{0}')".format(self.uri)

    def __hash__(self, *args, **kwargs) -> int:
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Sid):
            return str(other.uri) == str(self.uri)
        else:
            return str(other) == str(self)

    def __lt__(self, other: Sid | str) -> bool:
        return str(self) < str(other)

    def __truediv__(self, other: Sid | str | None) -> Sid:
        """
        Override the division operator, to compose a Sid from another one.
        Inspired by pathlib.Path.

        The get_with() method is prefered, since it generally is less "volatile".

        Examples:

            >>> Sid('hamlet') / 's' / 'sq030'
            Sid('shot__sequence:hamlet/s/sq030')

            >>> sid = Sid("hamlet/s/sq030")
            >>> sid.parent / sid.get(sid.keytype) == sid
            True

            >>> Sid("hamlet/s/sq030") / None
            Sid('shot__sequence:hamlet/s/sq030')
        """
        if other is None:
            return self.copy()
        return Sid(str(self) + conf.sip + str(other))


class TypedSid(StringSid):
    """
    The TypedSid class implements a Sid that has been resolved successfully.

    It has a "type" and a "field" data dictionary.
    """

    def _init(
        self,
        string: Optional[str] = None,
        type: Optional[str] = None,
        fields: Optional[dict] = None,
    ):
        self._string = string or ""
        self._type = type or ""
        self._fields = fields or {}

    @property
    def type(self) -> str:
        """
        Returns the Sids type.

        If the Sid has no type, an empty string is returned.

        Examples:

            >>> Sid("hamlet/a/char/ophelia/rig").type
            'asset__task'

            >>> Sid("hamlet/a/char/*").type
            'asset__asset'

            >>> Sid("blabla").type
            ''

        Returns:
            The Sid type

        """
        return self._type

    @property
    def fields(self) -> dict:
        """
        Returns the Sids data dictionary.

        The fields match what is defined in the Sid template.

        Examples:

            >>> Sid("hamlet/a/char/*").fields
            OrderedDict([('project', 'hamlet'), ('type', 'a'), ('assettype', 'char'), ('asset', '*')])

            >>> Sid("hamlet/s/sq010/sh0010/anim").fields
            OrderedDict([('project', 'hamlet'), ('type', 's'), ('sequence', 'sq010'), ('shot', 'sh0010'), ('task', 'anim')])

        Returns:
            Sids data dictionary.

        """
        return self._fields.copy()

    @property
    def uri(self) -> str:
        """
        Returns the "uri" representation of a Sid.

        The form is "type:string" if typed, else just the sid string.

        Examples:

            >>> Sid('hamlet/s/sq030/sh0100/anim').uri
            'shot__task:hamlet/s/sq030/sh0100/anim'

            >>> Sid('blablabla').uri
            'blablabla'

        Note:
             If the Sid ends with a "query", this query is applied, if possible.
             An unapplied query stays in the string and thus, the uri.

        Examples with query:

            >>> Sid('hamlet/s/sq030/sh0100/anim?task=render').uri
            'shot__task:hamlet/s/sq030/sh0100/render'

            >>> Sid('hamlet/s/sq030/sh0100/anim?foo=bar').uri
            'shot__task:hamlet/s/sq030/sh0100/anim?foo=bar'

            >>> Sid('blablabla?foo=bar').uri
            'blablabla?foo=bar'

        Returns
            the "uri" representation of a Sid.

        """
        return "{}{}".format(self._type + ":" if self._type else "", self.string)

    @property
    def basetype(self) -> str | None:
        """
        The basetype is the first part of the type.

        Examples:

            >>> Sid("hamlet/s/sq030/sh0100/anim").basetype
            'shot'

            >>> Sid("hamlet/a/prop/dagger").basetype
            'asset'

            >>> Sid("bla/bla").basetype

        Returns
            string basetype

        """
        result = None
        if not self._type:
            info(f"This Sid has no type. ({self})")
            return None
        try:
            result = self._type.split(conf.sidtype_keytype_sep)[0]
        except Exception as e:
            info(f'[sid][basetype] Unable to get basetype. Sid: {self} ("{e}")')
        return result

    @property
    def keytype(self) -> str | None:
        """
        Returns the last key of the fields dictionary.

        Note that this is not necessarily the second part of the "type".

        Note:

            Sid("hamlet/a/char/claudius/model/v001/w/blend")
              - type: 'asset__file'
              - keytype: 'ext'
              - basetype: 'asset'

            Sid("hamlet/s")
              - type: 'shot'
              - keytype: 'type'
              - basetype: 'shot'

            Sid("hamlet")
              - type: 'project'
              - keytype: 'project'
              - basetype: 'project'

        Examples:

            >>> Sid("hamlet/a/char/claudius/model/v001/w/blend").keytype
            'ext'

            >>> Sid("hamlet/a/char/ophelia/model").keytype
            'task'

            >>> Sid("hamlet/s/sq001/sh0010").keytype
            'shot'

        Returns:
             string keytype

        """
        if not self._fields:
            warning(f'Sid operation on an undefined Sid "{self.string}"')
            return None
        return list(self._fields.keys() or [None])[-1]

    @property
    def parent(self) -> Sid:
        """
        Returns the parent Sid.

        Returns an empty Sid, if the Sid is not "defined", or a copy of self if the Sid is already the root (has no parent).

        Note that this is a logical operation, without data access.

        Examples:

            >>> Sid('hamlet/s/sq030/sh0100').parent
            Sid('shot__sequence:hamlet/s/sq030')

            >>> Sid('hamlet').parent
            Sid('project:hamlet')

            >>> Sid('bla/bla/bla').parent
            Sid('')

        Returns:
            Returns an empty Sid, if the Sid is not "defined", or self if the Sid is already the root (has no parent).

        """
        if not self._fields:
            warning(f'Sid operation on an undefined Sid "{self.string}"')
            return Sid()
        if len(self._fields.keys()) == 1:
            return self.copy()
        parent_key = list(self._fields.keys())[-2]
        return self.get_as(parent_key)

    def __len__(self) -> int:
        """
        Returns the "length" of a Sid, which is the number of keys in its "fields" data dictionary.

        Returns:
            The amount of keys in the "fields" dictionary.

        """
        return len(self._fields)

    def get(self, key: str) -> str | None:
        """
        Returns the value of given key.
        As defined in the internal "fields" data dictionary.

        Example:

            >>> Sid("hamlet/s/sq030/sh0100/anim").get("shot")
            'sh0100'

        Args:
            key: retrieved key

        Returns:
            Value of given key, in the "fields" data dictionary.

        """
        if not self._fields:
            warning(f'Sid operation on an undefined Sid "{self.string}"')
            return None

        return self._fields.get(key)

    def get_as(self, key: str) -> Sid:
        """
        Returns a new Sid built of the fields until (and including) the given key.

        If the Sid is not typed or the key is not found, an empty Sid is returned.

        Examples:

            >>> Sid('hamlet/a/char/ophelia/model/v001/w/ma').get_as('task')
            Sid('asset__task:hamlet/a/char/ophelia/model')

            >>> Sid('hamlet/a/char/ophelia/model/v001/w/ma').get_as('something')
            Sid('')

            >>> Sid('hamlet/bla/bla').get_as('shot')
            Sid('')

        Args:
            key: key until which the new Sid should be built

        Returns:
            Sid as given key

        """
        if not self._fields:
            warning(f'Sid operation on an undefined Sid "{self.string}"')
            return Sid()  # Return type is always Sid.

        if key not in self._fields:
            info(f'Key "{key}" not found in fields "{self._fields}"')
            return Sid()

        fields = {}
        for k, v in self._fields.items():
            fields[k] = v
            if k == key:
                return Sid(fields=fields)

        raise SpilException(f"[Sid][get_as] Unexpected error during {self}.get_as({key})")

    def get_with(
        self,
        query: Optional[str] = None,
        key: Optional[str] = None,
        value: Optional[str] = None,
        **kwargs,
    ) -> Sid:
        """
        Returns a new Sid.

        - with the given query applied (see details about query application in documentation)

        or

        - updated using given key and value, eg. get_with(key='task', value='rendering') and
        - updated using **kwargs where keys are sid keys, eg. get_with(task='rendering')

        Depending on the update, the type of the returned Sid can change, or the Sid can end up untyped.

        A key set to None will be removed.
        To empty a keys value, set it to an empty string "".

        Note:
            Only works on typed or empty Sids.

        Examples:

            >>> Sid('hamlet/s/sq030/sh0010/anim').get_with(task='render')
            Sid('shot__task:hamlet/s/sq030/sh0010/render')

            >>> Sid("hamlet/a/prop/dagger").get_with(query='asset=skull')
            Sid('asset__asset:hamlet/a/prop/skull')

            >>> Sid().get_with(project="hamlet")
            Sid('project:hamlet')

        Args:
            query: s tring query, in the format ?key=value&key2=value2
            key: a key name, for example "shot", "sequence", or "task"
            value: a new value for given key. May be None (removes the key) or "" (empty value).
            **kwargs: a key/value dictionary

        Returns:
            A new Sid. Depending on the update, the type of the returned Sid can change.

        """
        if self._string and not self._fields:
            warning(f'Sid operation on an undefined Sid "{self.string}"')
            return Sid()  # Return type is always Sid.

        # if we have a query
        if query:
            return Sid("{}?{}".format(self.uri, query))

        data_copy = self._fields.copy()

        if key:
            kwargs[key] = value

        for key, value in kwargs.copy().items():
            # removing a key if the value is None. Use '' for empty values.
            if value is None:
                data_copy.pop(key)
                kwargs.pop(key)

        data_copy.update(kwargs)
        new_sid = Sid(fields=data_copy)
        # If the resulting Sid is not typed, and is a search, we try return as StringSid, matching the Sid dictionary.
        # This is useful but may have unexpected side effects.
        if new_sid.is_search() and not new_sid:
            new_sid = Sid("/".join(list(data_copy.values())))
        return new_sid

    def is_leaf(self):
        """
        Returns True if the current Sid is a "leaf" node.

        A leaf node has no children (and cannot have children).

        Current implementation checks if the keytype is set as a "leaf type" in the config_name.
        For example: "ext" (file extension) is the keytype of leaf nodes.
        This is configured per basetype.

        This implementation is still experimental.
        Implementation and concept needs to be clarified.

        Note: A leaf should be dependent on the context and type.
        For example, in searching for render files,
        it can be useful to handle the "render pass" as leave,
        to avoid going too deep in the hierarchy.

        This is done, for example, in the spil_ui browser,
        to browse render files by the pass folder,
        not individually by default.

        Returns:
            True if this Sid is a leaf, else False.

        """
        return bool(self.get(conf.leaf_keys.get(self.basetype)))
        # TODO:
        # Better define "complete". Also in regard to a search Sid.
        # For example Sids containing /** are "complete".
        # or if a Sid has no children, it is leaf.

    def as_query(self) -> str:
        """
        Returns the fields as a key value string, as in an Query.

        Example:

            >>> Sid('hamlet/a/char/ophelia/model').as_query()
            'project=hamlet&type=a&assettype=char&asset=ophelia&task=model'

        Returns:
            string query

        """
        return query_helper.to_string(self._fields)

    # IDEA: match_as(search_sid, key) for example, do the "seq" of both sids match (like is_relative_to ?)
    def match(self, search_sid: Sid | str) -> bool:
        """
        Returns True if a given search_sid matches the current Sid, else False.

        This method is useful to create filters matching groups of Sids, more precise than types.

        Examples:

            >>> Sid('hamlet/a/char/ophelia/model').match('hamlet/a/*/*/model')
            True

            >>> Sid('hamlet/s/sq030/sh0100/anim').match('hamlet/s/*/*/*')
            True

            >>> Sid('hamlet/a/char/ophelia').match('hamlet/a/prop/*')
            False

        Returns
            True if matched, else False

        """
        # Identicals always match
        if Sid(search_sid) == self:
            return True
        # Should untyped sids be able to match ? Identical strings could match.
        if not self._fields:
            warning(f'Cannot match check an undefined Sid: "{self.string}". Returning False.')
            return False
        # To check the match, we search in a list with self as single element,
        # and expect self to be found using given search_sid.
        from spil import FindInList  # fmt: skip
        fl = FindInList([self.string])
        return fl.find_one(search_sid, as_sid=False) == self.string


class PathSid(TypedSid):
    """
    The PathSid adds path resolving.

    TODO: cache must be invalidated whenever (if ever) the default PathConfig changes.
    """

    @cache
    def path(self, config: Optional[str] = None) -> Path | None:
        """
        Returns the file path for the current Sid, as a pathlib.Path.
        Returns None if the Sid has no path, or if it cannot be resolved.

        This method is cached.
        A path for a Sid is configured (not the result of a data lookup),
        and is not supposed to change during runtime.

        Example:

            >>> sid = Sid('hamlet/a/char/ophelia/model/v001/w/ma')
            >>> path = sid.path()
            >>> new_sid = Sid(path=path)
            >>> sid == new_sid
            True

            >>> path = Sid('hamlet/a/char/ophelia/model/v001/w/ma').path()
            >>> path.relative_to(conf.default_sid_conf_data_path).as_posix()  # to be location and os independent
            'testing/SPIL_PROJECTS/LOCAL/PROJECTS/HAMLET/PROD/ASSETS/char/ophelia/model/v001/char_ophelia_model_WORK_v001.ma'

            >>> Sid('bla/bla').path()

        Args:
            config: Name of the path config to be used, as configured.

        Returns:
            A path, if the Sid has a path, else None.
        """
        if not self._fields:
            debug(f'Sid is undefined: "{self.string}". Returning None.')
            return None
        from spil.sid.pathops.fs_resolver import dict_to_path  # fmt: skip
        result = None
        try:
            result = dict_to_path(self._fields, self._type, config=config)
        except SpilException as e:
            debug(f"This Sid has no path (config_name: {config}). ({e})")
        return result


class DataSid(PathSid):
    """
    The DataSid implements operations that delegate calls to data sources: Finders and Getters.
    By default FindInAll and GetFromAll are used.

    # TODO: make configurable which Finder is used for DataSid operations (FindInAll per default).
    # Could be handled using a default config_name, and/or be changed at runtime.
    """

    def get_attr(self, attribute: str) -> Any | None:
        """
        Returns an attribute for the current sid.
        Uses GetFromAll, which calls the apropriate Getter defined for this Sid in spil_data_conf.

        Shortcut to GetFromAll().get_attr(sid, attribute), which is called internally.

        Example:

            >>> from spil import WriteToPaths
            >>> sid = Sid('hamlet/a/char/ophelia/model/v001/w/ma')
            >>> __ = WriteToPaths().set(sid, comment="Updated topology")
            >>> sid.get_attr('comment')
            'Updated topology'

        Args:
            attribute: Name of an attribute

        Returns:
            The value of the attribute, or None if it was not found.

        """
        from spil import GetFromAll  # fmt: skip
        return GetFromAll().get_attr(self, attribute)  # type: ignore

    def get_last(self, key: Optional[str] = None) -> Sid:
        """
        Returns a new Sid object, with the same fields as self, and the last existing match for given key.
        If key is not given, the keytype is used.
        If the Sid is untyped, returns an empty Sid.

        Examples:

            >>> Sid('hamlet/a/char/ophelia/model/v001/p/ma').get_last('version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v002/p/ma')

            >>> Sid('hamlet/a/char/ophelia/model').get_last()
            Sid('asset__task:hamlet/a/char/ophelia/surface')

            >>> Sid('bla/bla').get_last('whatever')
            Sid('')

        This method calls FindinAll internally, and is a shortcut to:
            FindInAll().find_one(self.get_with(key=key, value='>'), as_sid=True)

        Note: Sid sorting is currently limited to string values, which usually works with versions.
        A meaningful sorting of tasks (eg. "render" after "animation") is planned.

        (implementation is experimental)

        Returns:
            Sid or None
        """
        if not self._fields:
            debug(f'Sid is undefined: "{self.string}". Returning Emptu Sid..')
            return Sid()
        if not key:
            key = self.keytype
        from spil import FindInAll  # fmt: skip
        found = FindInAll().find_one(self.get_with(key=key, value=">"), as_sid=True)
        if found.get(key):  # little failsafe. #SMELL
            return found
        else:
            return Sid()

    def get_next(self, key: str) -> Sid:
        """
        Returns self with key's value incremented, or first value if there is none.
        If value is '*', returns "get_new" (next of last)

        If the result is not a valid Sid (not typed, no fields), returns an empty Sid.

        Note:
            Currently limited to version.

        Examples:

            >>> Sid('hamlet/a/char/ophelia/model/v001/w/ma').get_next('version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v002/w/ma')

            >>> Sid('hamlet/a/char/ophelia/model/*/w/ma').get_next('version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v003/w/ma')

            >>> Sid('hamlet/a/char/ophelia/model').get_next('version')
            Sid('asset__version:hamlet/a/char/ophelia/model/v001')

        Args:
            key: the key for which we want to increment the value. Typically a version.

        Returns:
            A Sid with key's value incremented, or an empty Sid.

        """
        if key != "version":
            raise NotImplementedError("get_next() support only 'version' key for the moment.")

        attribute = f"next.{key}"

        from spil import GetFromAll  # fmt: skip
        return GetFromAll().get_attr(self, attribute=attribute)  # type: ignore

    def get_new(self, key: str) -> Sid:  # FIXME: Needs testing and documentation
        """
        Returns a new Sid with the key incremented to the "next available value".
        Makes sense with numerical values, especially with versions.

        For example, if the given key is "version":
        Returns self with next version (the one following the last), or first version if there is no version.
        If the result is not a valid Sid (not typed, no fields), returns an empty Sid.

        Example:

            >>> Sid('hamlet/a/char/ophelia/model/v001/w/ma').get_new('version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v003/w/ma')

        Args:
            key:

        Returns:
            Sid

        """
        if self.get(key):
            if self.get_last(key):
                result = self.get_last(key).get_next(key)
                return result or Sid()
            else:
                result = self.get_next(key)  # Returns a first version
                return result or Sid()
        else:
            with_added_key = self.get_with(key=key, value="*").get_last(key)
            if with_added_key:
                result = with_added_key.get_next(key)
                return result or Sid()
            else:
                result = self.get_next(key)  # Returns a first version
                return result or Sid()

    def exists(self) -> bool:
        """
        Returns True if the current Sid exists.
        Shortcut to FindInAll().exists(self), which is called internally.

        Examples:

            >>> Sid('hamlet/a/char/ophelia').exists()
            True

            >>> Sid('hamlet/a/prop/computer').exists()
            False

            >>> Sid('bla/bla').exists()
            False

        Returns:
            True if Sid exists, else False
        """
        if not self._fields:
            debug(f'Sid is undefined: "{self.string}". Returning False')
            return False
        from spil import FindInAll  # fmt: skip
        return FindInAll().exists(self)  # type: ignore

    def siblings_as(self, key: str) -> List[Sid]:
        """
        Returns existing siblings, in other words children of parent, as Sid of given keytype.

        Example:

            >>> sorted( Sid('hamlet/a/fx/blood/surface/v002/p/ma').siblings_as('asset') )
            [Sid('asset__asset:hamlet/a/fx/blood'), Sid('asset__asset:hamlet/a/fx/mist'), Sid('asset__asset:hamlet/a/fx/rain'), Sid('asset__asset:hamlet/a/fx/thunder'), Sid('asset__asset:hamlet/a/fx/water')]

        Args:
            key: keytype we want siblings from

        Returns:
            List of sibling Sids

        """
        if key not in self._fields:
            info(f'[Sid][siblings_as] Key "{key}" not found in fields "{self._fields}"')
            return []
        search = self.get_as(key).get_with(key=key, value="*")
        from spil import FindInAll  # fmt: skip
        return list(FindInAll().find(search, as_sid=True))

    def siblings(self) -> List[Sid]:
        """
        Returns existing siblings, in other words children of parent.

        Example:

            >>> sorted( Sid('hamlet/a/fx/blood/surface').siblings() )
            [Sid('asset__task:hamlet/a/fx/blood/art'), Sid('asset__task:hamlet/a/fx/blood/model'), Sid('asset__task:hamlet/a/fx/blood/rig'), Sid('asset__task:hamlet/a/fx/blood/surface')]

        Returns:
            List of sibling Sids.
        """
        return self.siblings_as(self.keytype)

    def children(self) -> List[Sid]:
        """
        Returns existing children.
        Shortcut to:  `FindInAll().find(self / '*', as_sid=True)`

        Examples:

            >>> Sid('hamlet/a/char/ophelia/rig').children()
            [Sid('asset__version:hamlet/a/char/ophelia/rig/v001'), Sid('asset__version:hamlet/a/char/ophelia/rig/v002')]

            >>> Sid('hamlet/a/char/ophelia/model/v001/w/ma').children()
            []

        Returns:
            Children Sids
        """
        if self.is_leaf():  # per definition, leafs have no children.
            return []
        from spil import FindInAll  # fmt: skip
        search = self / "*"
        return list(FindInAll().find(search, as_sid=True))

    # def get_first, get_previous,
    # define "complete / incomplete" sid.is_leaf ? (root / anchor / parent / parents )


class Sid(DataSid):
    """
    Sid class.

    Multiple ways to create a Sid:
    - Sid string, uri, query string (starting with "?"), Sid object, or empty string
    - query string
    - fields dictionary
    - path (with optional config)

    Examples:

        >>> Sid("hamlet/s/sq010")                                   # string
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid(sid="hamlet/s/sq010")                               # parameter + string
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid(Sid("hamlet/s/sq010"))                              # Sid object
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid("shot__sequence:hamlet/s/sq010")                    # uri
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid("?project=hamlet&type=s&sequence=sq010")             # empty string with query
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid(query="project=hamlet&type=s&sequence=sq010")          # query
        Sid('shot__sequence:hamlet/s/sq010')

        >>> Sid(fields={'project': 'hamlet', 'sequence': 'sq010', 'type': 's'})  # fields dict
        Sid('shot__sequence:hamlet/s/sq010')

        >>> path = Path(conf.default_sid_conf_data_path) / "testing/SPIL_PROJECTS/LOCAL/PROJECTS/HAMLET/PROD/ASSETS/char/ophelia/model/v001/char_ophelia_model_WORK_v001.ma"
        >>> Sid(path=path)           # path (default config) # TODO: any config
        Sid('asset__file:hamlet/a/char/ophelia/model/v001/w/ma')

        >>> Sid(path=path, config='local')   # path (config "local")
        Sid('asset__file:hamlet/a/char/ophelia/model/v001/w/ma')

    This class inherits all functionality from the class hierarchy.
    It only defines the factory to be used to create the Sid.

    Note that this implementation may change, but API will remain the same.
    The goal is to be able to Extend the Sid class to custom needs.
    """

    _factory = ("spil.sid.core.sid_factory", "sid_factory")  # TODO: config_name, or better system.


if __name__ == "__main__":

    # from spil.util.log import setLevel, INFO
    # setLevel(INFO)

    s = Sid("hamlet/s/sq010/sh0010/anim")
    print(s.exists())
    print(list(s.children()))



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

@author: michael.haussmann
"""


class Singleton(object):

    instances = {}

    def __new__(cls, *args, **kwargs):
        if Singleton.instances.get(cls) is None:
            # Singleton.instances[cls] = object.__new__(cls, *args, **kwargs)
            Singleton.instances[cls] = object.__new__(cls)
        return Singleton.instances[cls]


if __name__ == '__main__':

    class Person(Singleton):

        def __init__(self, name):
            self.name = name

        def __str__(self):
            return '[Person] : {}'.format(self.name)

    john = Person('john')
    print(john)

    jane = Person('jane')
    print(jane)

    assert(john is jane)

    print(john)
    print(jane)




"""
Test if the given Sid list covers all the templates.

This allows to make sure the configuration test is complete.

"""
from pprint import pformat

from spil import Sid

from spil.conf import sid_templates  # type: ignore

from spil.util.log import DEBUG, ERROR, get_logger
log = get_logger('tests')
log.setLevel(DEBUG)


def test_config_coverage(sids):

    all_types = set(sid_templates.keys())
    covered = set()
    remaining = all_types.copy()

    log.debug('- Testing coverage of given sids for sid_templates (sid_conf)...')

    for sid in sids:
        _sid = Sid(sid)
        if _sid.type in remaining:
            covered.add(_sid.type)
            log.debug(f"Covering: {_sid.type}")
            remaining.remove(_sid.type)
            if not remaining:
                break

    missing = remaining

    log.debug(f"Covered: {pformat(covered)}")

    if missing:
        log.warning(f"\tFAILED: Types not covered by given sids: {pformat(missing)}")
    else:
        log.info('\tOK: All types covered.')

    log.debug('done')


if __name__ == '__main__':

    sids = ["hamlet", "hamlet/a", "hamlet/a/char"]
    test_config_coverage(sids)



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""

# from tests.utils import init  # needs to be before spil.conf import

from spil.util.log import debug, info
from spil import Sid
from spil.conf import projects  # type: ignore

from pathlib import Path


def is_ok(path):
    """
    Quick filter function to speed up parsing.
    """
    if path.name.startswith('.') or path.name.startswith('_') or path.suffix.endswith('~'):
        return False
    if path.suffix in ['.png', '.exr', '.jpg', '.raysync', '.swatches', '.comment']:  # currently not supported - speed up parsing
        return False
    return True


def parse_sids_from_files(filter_callback=None, config=None):
    """
    Recursively traverses the project directories, fetches and yields all compatible Sids.

    config is the Spil path config to be used when instantiating Sids.
    If None, the default config is used.

    is_ok_callback is a function to filter out paths that do not need to be Sid tested.

    Args:
        filter_callback: function that returns False if the path should be ignored, else True.
        config:

    Returns:

    """
    if filter_callback:
        is_ok = filter_callback
    else:
        def is_ok(path):
            return True

    for project in projects[:]:

        project_root = Path(Sid(project).path(config))
        info('Root path : {}'.format(project_root))

        for path in project_root.rglob('*'):
            if not is_ok(path):
                continue
            debug(path)
            sid = Sid(path=path, config=config)
            if str(sid):
                yield sid


if __name__ == '__main__':

    from spil.util.log import setLevel, DEBUG, INFO, ERROR

    setLevel(ERROR)  # Set to ERROR if file system contains a lot of non Sid translatable files.

    generator = parse_sids_from_files(filter_callback=is_ok)

    for sid in generator:
        print(f"Found: {sid} \t @ \t {sid.path()}")



from spil.util.exception import raiser
from spil import Sid, Getter


class NextGetter(Getter):  # noqa (does not implement all of superclass)
    """
    Receives specific "next." attributes and returns accordingly.

    Currently implemented to calculate and retrieve the "next" version of a Sid.
    """

    def get_attr(self, sid, attribute):
        """
        Acts as a callback for Sid.get_next(), via GetFromAll.

        See Sid.get_next documentation.

            >>> NextGetter().get_attr('hamlet/a/char/ophelia/model/v001/w/ma', 'next.version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v002/w/ma')

            >>> NextGetter().get_attr('hamlet/a/char/ophelia/model/*/w/ma', 'next.version')
            Sid('asset__file:hamlet/a/char/ophelia/model/v003/w/ma')

            >>> NextGetter().get_attr('hamlet/a/char/ophelia/model', 'next.version')
            Sid('asset__version:hamlet/a/char/ophelia/model/v001')

        Args:
            sid:
            attribute:

        Returns:

        """

        spec, key = attribute.split(".")
        (spec == "next" and key == "version") or raiser("This implementation is limited to returning the next version Sid.")

        _sid = Sid(sid)
        current = _sid.get("version")

        # extracts current version
        if current:
            if current in ["*", ">"]:
                version = (_sid.get_last("version").get("version") or "v000").split("v")[-1] or 0
            else:
                # temporary workaround for "v001"
                version = str(current).split("v")[-1]
        # or 0 if no current version
        else:
            version = 0  # starts with v001

        # increments and formats
        version = int(version) + 1
        version = "v" + str("%03d" % version)

        # builds Sid and returns
        result = _sid.get_with(version=version) or Sid()
        return result



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil import Sid, FindInPaths, SpilException
from spil.tests.utils.sid_full_tester import test_full_sid
from spil.util.log import get_logger
from spil.tests import Timer

log = get_logger("spil_tests")

"""
To launch searches on a custom Finder.

Searches is a dict with searches as 
    key: search sid
    value: search description 

"""

def check_searches_in_finder(searches, finder=None, as_sid=True, do_log=True, do_deep=False, do_doublon_check=True, do_match_check=False, replace=None, reraise=True):
    """
    Runs given searches on the given Finder.

    Optionally operates a replace in the read, using given replace tuple.
    """

    global_timer = Timer(name="global", logger=log.info)
    global_timer.start()

    finder = finder or FindInPaths()
    for search_sid, comment in searches.items():

        log.info("*" * 10)
        log.info("{} --> {} (finder: {})".format(search_sid, comment, finder))
        double_check = set()

        try:
            f_timer = Timer(name="search_sid", logger=log.info)
            f_timer.start()
            count = 0
            for i in finder.find(search_sid, as_sid=as_sid):
                if do_log:
                    log.info(i)
                if do_match_check:
                    match = Sid(i).match(search_sid)
                    if not match:
                        log.warning('No match "{}" <-> "{}". This is not normal'.format(i, search_sid))
                count += 1
                if do_doublon_check:
                    if i in double_check:
                        log.warning("--------------------------------------> Doublon {}".format(i))
                    double_check.add(i)
                if do_deep:
                    log.debug("Sid test for {}".format(i))
                    test_full_sid(i, from_search=search_sid)

            log.info("Total: " + str(count))
            f_timer.stop()

        except SpilException as e:
            log.error("SpilException : {} --> {}".format(e, search_sid))
            if reraise:
                raise e

    global_timer.stop()


if __name__ == "__main__":

    from spil.util.log import setLevel, ERROR

    setLevel(ERROR)

    searches = {}
    searches["hamlet/s/*/*"] = ""

    check_searches_in_finder(searches)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, Mapping, Any, List, Optional, Callable

from spil import Sid, Getter, Finder
from spil.conf import default_path_config, path_data_suffix  # type: ignore


class GetByFinder(Getter):
    """
    Some Getters first use a Finder to retrieve the Sids,
    then, in a second distinct step, retrieve the data.
    For example, GetFromPaths which stores data in json "sidecar" files.

    This "GetByFinder" is an abstract Getter, implementing shared methods for Getters that use a Finder.

    These Getters can inherit GetByFinder, and:
    - override the constructor and define self.finder
    - or pass it to the constructor (inherited from GetByFinder)
    - and implement get_data()
    """

    def __init__(self, finder: Finder):
        self.finder = finder

    def get(
        self,
        search_sid: str | Sid,
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Iterator[Mapping[str, Any]]:
        """
        See Getter.
        """
        for sid in self.finder.find(search_sid=search_sid, as_sid=True):
            data = self.get_data(sid, attributes=attributes, sid_encode=sid_encode) or {}
            yield data

    def do_get(
        self,
        search_sids: List[Sid],
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Iterator[Mapping[str, Any]]:
        """
        See Getter.
        """
        for sid in self.finder.do_find(search_sids=search_sids, as_sid=True):
            data = self.get_data(sid, attributes=attributes, sid_encode=sid_encode) or {}
            yield data

    def get_data(
        self, sid: str | Sid, attributes: Optional[List[str]] = None, sid_encode: Callable = str
    ) -> Mapping[str, Any]:
        """
        See Getter.
        """
        raise NotImplementedError("To be used, GetByFinder should be inherited, and get_data overriden.")

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Finder: "{self.finder}"]'


if __name__ == "__main__":
    from spil import FindInPaths

    finder = FindInPaths()
    print(GetByFinder(finder))



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.util.exception import raiser
from spil.util.log import get_logger
from spil.tests import Timer

log = get_logger("spil_tests")

"""
To launch searches on two finders. 
Allows comparing of finders.

Searches is a dict with searches as 
    key: search sid
    value: search description 

"""

def check_search_ab(searches, finderA, finderB, as_sid=False, raise_problems=False):

    log.debug("Searching sids in A:{} and B:{}".format(finderA, finderB))

    global_timer = Timer(name="global", logger=log.debug)
    global_timer.start()

    for search_sid, comment in searches.items():

        log.info("=" * 50)
        log.info("{} --> {}".format(search_sid, comment))

        a_timer = Timer(name="a_timer", logger=log.info)
        a_timer.start()
        found_a = set(finderA.find(search_sid, as_sid=as_sid))
        log.info("A : {}".format(len(found_a)))
        a_timer.stop()

        b_timer = Timer(name="b_timer", logger=log.info)
        b_timer.start()
        found_b = set(finderB.find(search_sid, as_sid=as_sid))
        log.info("B : {}".format(len(found_b)))
        b_timer.stop()

        problems = found_b ^ found_a

        for i in problems:
            log.warning(f"Not equal in both searches: {i}")

        if problems and raise_problems:
            raiser(f"Differences: {problems}")

    log.debug("=" * 10)
    log.debug("Done all searches.")
    global_timer.stop()


if __name__ == "__main__":

    from spil import FindInPaths, FindInAll
    from spil.util.log import setLevel, ERROR

    setLevel(ERROR)

    searches = {}
    searches["*"] = "projects"
    searches["hamlet/*"] = "types"
    searches["hamlet/a/*"] = "asset types"
    searches["hamlet/a/*/*"] = "assets"

    fpl = FindInPaths('local')
    fps = FindInPaths('server')
    # fl = FindInList(sids)
    ft = FindInAll()

    check_search_ab(searches, ft, fpl)
    # to compare speed, run the test twice and check second time.
    # check_search_ab(searches, fpl, ft)



# type: ignore
"""
This script uses the test sids,
and generates empty but correctly named files,
according to the path configs.

"""
import spil_data_conf
from hamlet_scripts.example_sids import sids
from spil.tests.prep.make_mock_fs import generate_mock_fs

if __name__ == "__main__":

    from spil.util.log import setLevel, INFO

    print("")
    print("Tests start")
    print("")

    # setLevel(WARN)
    setLevel(INFO)
    # setLevel(DEBUG)  # In case of problems, use DEBUG mode

    for config in list(spil_data_conf.path_configs):
        generate_mock_fs(sids, config=config)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.conf.global_conf import __version__

try:
    from spil import conf  # default config bootstrap
    from spil.sid.sid import Sid

    from spil.sid.read.finder import Finder
    from spil.sid.pathops.find_paths import FindInPaths
    from spil.sid.read.finders.find_list import FindInList
    from spil.sid.read.finders.find_constants import FindInConstants
    from spil.sid.read.finders.find_all import FindInAll
    from spil.sid.read.getters.getter_all import GetFromAll
#     from spil.sid.read.finders.find_cache import FindInCache

    from spil.sid.read.getter import Getter
    from spil.sid.pathops.getter_paths import GetFromPaths

    from spil.sid.write.writer import Writer
    from spil.sid.write.write_all import WriteToAll
    from spil.sid.pathops.write_paths import WriteToPaths

    # FIXME: keep util level import
    from spil.util.exception import SpilException
    from spil.util import log
    from spil.util import log as logging  # to use as standard logging and create custom loggers
    from spil.util.log import setLevel, ERROR

    setLevel(ERROR)
except Exception as e:
    import traceback  # fmt: skip
    traceback.print_exc()
    raise Exception(
        "Spil is imported, but impossible to import spil packages. \n Please check compatibility of your sid_conf and fs_conf files."
    )



# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

import abc
import sys
import re
import functools
from collections import defaultdict
from future.utils import with_metaclass

import lucidity.error

# Type of a RegexObject for isinstance check.
_RegexType = type(re.compile(''))


class Template(object):
    '''A template.'''

    _STRIP_EXPRESSION_REGEX = re.compile(r'{(.+?)(:(\\}|.)+?)}')
    _PLAIN_PLACEHOLDER_REGEX = re.compile(r'{(.+?)}')
    _TEMPLATE_REFERENCE_REGEX = re.compile(r'{@(?P<reference>.+?)}')

    ANCHOR_START, ANCHOR_END, ANCHOR_BOTH = (1, 2, 3)

    RELAXED, STRICT = (1, 2)

    def __init__(self, name, pattern, anchor=ANCHOR_START,
                 default_placeholder_expression='[\w_.\-]+',
                 duplicate_placeholder_mode=RELAXED,
                 template_resolver=None):
        '''Initialise with *name* and *pattern*.

        *anchor* determines how the pattern is anchored during a parse. A
        value of :attr:`~Template.ANCHOR_START` (the default) will match the
        pattern against the start of a path. :attr:`~Template.ANCHOR_END` will
        match against the end of a path. To anchor at both the start and end
        (a full path match) use :attr:`~Template.ANCHOR_BOTH`. Finally,
        ``None`` will try to match the pattern once anywhere in the path.

        *duplicate_placeholder_mode* determines how duplicate placeholders will
        be handled during parsing. :attr:`~Template.RELAXED` mode extracts the
        last matching value without checking the other values.
        :attr:`~Template.STRICT` mode ensures that all duplicate placeholders
        extract the same value and raises :exc:`~lucidity.error.ParseError` if
        they do not.

        If *template_resolver* is supplied, use it to resolve any template
        references in the *pattern* during operations. It should conform to the
        :class:`Resolver` interface. It can be changed at any time on the
        instance to affect future operations.

        '''
        super(Template, self).__init__()
        self.duplicate_placeholder_mode = duplicate_placeholder_mode
        self.template_resolver = template_resolver

        self._default_placeholder_expression = default_placeholder_expression
        self._period_code = '_LPD_'
        self._at_code = '_WXV_'
        self._name = name
        self._pattern = pattern
        self._anchor = anchor

        # Check that supplied pattern is valid and able to be compiled.
        self._construct_regular_expression(self.pattern)

    def __repr__(self):
        '''Return unambiguous representation of template.'''
        return '{0}(name={1!r}, pattern={2!r})'.format(
            self.__class__.__name__, self.name, self.pattern
        )

    @property
    def name(self):
        '''Return name of template.'''
        return self._name

    @property
    def pattern(self):
        '''Return template pattern.'''
        return self._pattern

    def expanded_pattern(self):
        '''Return pattern with all referenced templates expanded recursively.

        Raise :exc:`lucidity.error.ResolveError` if pattern contains a reference
        that cannot be resolved by currently set template_resolver.

        '''
        return self._TEMPLATE_REFERENCE_REGEX.sub(
            self._expand_reference, self.pattern
        )

    def _expand_reference(self, match):
        '''Expand reference represented by *match*.'''
        reference = match.group('reference')

        if self.template_resolver is None:
            raise lucidity.error.ResolveError(
                'Failed to resolve reference {0!r} as no template resolver set.'
                .format(reference)
            )

        template = self.template_resolver.get(reference)
        if template is None:
            raise lucidity.error.ResolveError(
                'Failed to resolve reference {0!r} using template resolver.'
                .format(reference)
            )

        return template.expanded_pattern()

    def parse(self, path):
        '''Return dictionary of data extracted from *path* using this template.

        Raise :py:class:`~lucidity.error.ParseError` if *path* is not
        parsable by this template.

        '''
        # Construct regular expression for expanded pattern.
        regex = self._construct_regular_expression(self.expanded_pattern())

        # Parse.
        parsed = {}

        match = regex.search(path)
        if match:
            data = {}
            for key, value in sorted(match.groupdict().items()):
                # Strip number that was added to make group name unique.
                key = key[:-3]

                # If strict mode enabled for duplicate placeholders, ensure that
                # all duplicate placeholders extract the same value.
                if self.duplicate_placeholder_mode == self.STRICT:
                    if key in parsed:
                        if parsed[key] != value:
                            raise lucidity.error.ParseError(
                                'Different extracted values for placeholder '
                                '{0!r} detected. Values were {1!r} and {2!r}.'
                                .format(key, parsed[key], value)
                            )
                    else:
                        parsed[key] = value

                # Expand dot notation keys into nested dictionaries.
                target = data

                parts = key.split(self._period_code)
                for part in parts[:-1]:
                    target = target.setdefault(part, {})

                target[parts[-1]] = value

            return data

        else:
            raise lucidity.error.ParseError(
                'Path {0!r} did not match template pattern.'.format(path)
            )

    def format(self, data):
        '''Return a path formatted by applying *data* to this template.

        Raise :py:class:`~lucidity.error.FormatError` if *data* does not
        supply enough information to fill the template fields.

        '''

        format_specification = self._construct_format_specification(
            self.expanded_pattern()
        )

        return self._PLAIN_PLACEHOLDER_REGEX.sub(
            functools.partial(self._format, data=data),
            format_specification
        )

    def _format(self, match, data):
        '''Return value from data for *match*.'''
        placeholder = match.group(1)
        parts = placeholder.split('.')

        try:
            value = data
            for part in parts:
                value = value[part]

        except (TypeError, KeyError):
            raise lucidity.error.FormatError(
                'Could not format data {0!r} due to missing key {1!r}.'
                .format(data, placeholder)
            )

        else:
            return value

    def keys(self):
        '''Return unique set of placeholders in pattern.'''
        format_specification = self._construct_format_specification(
            self.expanded_pattern()
        )
        return set(self._PLAIN_PLACEHOLDER_REGEX.findall(format_specification))

    def references(self):
        '''Return unique set of referenced templates in pattern.'''
        format_specification = self._construct_format_specification(
            self.pattern
        )
        return set(self._TEMPLATE_REFERENCE_REGEX.findall(format_specification))

    def _construct_format_specification(self, pattern):
        '''Return format specification from *pattern*.'''
        return self._STRIP_EXPRESSION_REGEX.sub('{\g<1>}', pattern)

    def _construct_regular_expression(self, pattern):
        '''Return a regular expression to represent *pattern*.'''
        # Escape non-placeholder components.
        expression = re.sub(
            r'(?P<placeholder>{(.+?)(:(\\}|.)+?)?})|(?P<other>.+?)',
            self._escape,
            pattern
        )

        # Replace placeholders with regex pattern.
        expression = re.sub(
            r'{(?P<placeholder>.+?)(:(?P<expression>(\\}|.)+?))?}',
            functools.partial(
                self._convert, placeholder_count=defaultdict(int)
            ),
            expression
        )

        if self._anchor is not None:
            if bool(self._anchor & self.ANCHOR_START):
                expression = '^{0}'.format(expression)

            if bool(self._anchor & self.ANCHOR_END):
                expression = '{0}$'.format(expression)

        # Compile expression.
        try:
            compiled = re.compile(expression)
        except re.error as error:
            if any([
                'bad group name' in str(error),
                'bad character in group name' in str(error)
            ]):
                raise ValueError('Placeholder name contains invalid '
                                 'characters.')
            else:
                _, value, traceback = sys.exc_info()
                message = 'Invalid pattern: {0}| {1}'.format(value, traceback)
                raise ValueError(message)

        return compiled

    def _convert(self, match, placeholder_count):
        '''Return a regular expression to represent *match*.

        *placeholder_count* should be a `defaultdict(int)` that will be used to
        store counts of unique placeholder names.

        '''
        placeholder_name = match.group('placeholder')

        # Support at symbol (@) as referenced template indicator. Currently,
        # this symbol not a valid character for a group name in the standard
        # Python regex library. Rather than rewrite or monkey patch the library
        # work around the restriction with a unique identifier.
        placeholder_name = placeholder_name.replace('@', self._at_code)

        # Support period (.) as nested key indicator. Currently, a period is
        # not a valid character for a group name in the standard Python regex
        # library. Rather than rewrite or monkey patch the library work around
        # the restriction with a unique identifier.
        placeholder_name = placeholder_name.replace('.', self._period_code)

        # The re module does not support duplicate group names. To support
        # duplicate placeholder names in templates add a unique count to the
        # regular expression group name and strip it later during parse.
        placeholder_count[placeholder_name] += 1
        placeholder_name += '{0:03d}'.format(
            placeholder_count[placeholder_name]
        )

        expression = match.group('expression')
        if expression is None:
            expression = self._default_placeholder_expression

        # Un-escape potentially escaped characters in expression.
        expression = expression.replace('\\{', '{').replace('\\}', '}')

        return r'(?P<{0}>{1})'.format(placeholder_name, expression)

    def _escape(self, match):
        '''Escape matched 'other' group value.'''
        groups = match.groupdict()
        if groups['other'] is not None:
            return re.escape(groups['other'])

        return groups['placeholder']


class Resolver(with_metaclass(abc.ABCMeta)):
    '''Template resolver interface.'''


    @abc.abstractmethod
    def get(self, template_name, default=None):
        '''Return template that matches *template_name*.

        If no template matches then return *default*.

        '''
        return default

    @classmethod
    def __subclasshook__(cls, subclass):
        '''Return whether *subclass* fulfils this interface.'''
        if cls is Resolver:
            return callable(getattr(subclass, 'get', None))

        return NotImplemented



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
from pathlib import Path

pysep = "/"  # python path separator

__version__ = "0.1.0"
application_codename = "Artichoke"
application_name = 'SPIL The Simple Pipeline Lib - v{0} ("{1}")'.format(
    __version__, application_codename
)

application_path = Path(__file__).parent.parent
default_sid_conf_path = str(application_path.parent / "spil_hamlet_conf")
default_sid_conf_data_path = str(application_path.parent / "spil_hamlet_conf" / "data")

# automatic replacement to INFO and WARN in BETA and PROD
loglevel = 80

user_app_folder_name = ".spil/conf"
user_conf_file_name = "user_conf.json"

# TODO: move these into sid_conf
sip = "/"  # sid separator - changing this is untested.
ors = ","  # "or" separator
qms = "#"  # question mark read sign (previously '?', thus it's name)
sidtype_keytype_sep = "__"
search_symbols = ["*", ",", ">", "<", "**"]

sid_conf_using_demo_configuration_message = f"""
-------------------------------------------------------------------------------------------------------------
USING DEMO CONFIGURATION

Spil needs configuration files in the pythonpath.
("spil_sid_conf.py" and others) 

None were found, so Spil is falling back to the shipped demo configuration.
The demo configuration is located here: "{default_sid_conf_path}".

You may adapt this demo configuration to your needs, or create a configuration on your own. 
Once you add your configuration folder to the pythonpath, this message will disappear.

Please see the documentation : https://spil.readthedocs.io
------------------------------------------------------------------------------------------------------------- 
"""

sid_conf_import_error_message = """
    -------------------------------------------------------------------------------------------------------------
    CONFIGURATION PROBLEM: 

    The configuration module "{module}" was not found.

    Ensure to either include the demo "spil_hamlet_conf" in your python path, 
    or create your own "{module}" and add its folder to the python path.    

    (If you are running a py.test edit the SPIL_CONF_PATH variable in tests/test_00_init.py to match a python path.)

    Please see installation and configuration documentation.

    -------------------------------------------------------------------------------------------------------------
    """



from spil.util.caching import lru_cache


def fib(n):
    if n in (1, 2):
        return 1
    else:
        return fib(n - 2) + fib(n - 1)

@lru_cache
def fib2(n):
    if n in (1, 2):
        return 1
    else:
        return fib2(n - 2) + fib2(n - 1)


from datetime import datetime


def time_fn(fn, *args):
    s = datetime.now()
    print(fn(*args))
    e = datetime.now()
    print(e - s)


if __name__ == '__main__':

    time_fn(fib, 40)
    time_fn(fib2, 40)


"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.sid.core.utils import expand


def execute(sids):
    """
    Expands multi-search characters like /** into a series of /*.
    The ** must exist only once.
    If values are present after the ** they are supposed to be the end of the sids.

    :param sids:
    :return:
    """
    result = []
    for sid in sids:
        result.extend(expand(sid, do_extrapolate=False))

    return result


if __name__ == "__main__":

    from pprint import pprint
    from spil.util.log import setLevel, info, INFO, debug, DEBUG

    info("Tests start")

    setLevel(INFO)

    expandables = [
        "hamlet/s/*/**/ma",
        "hamlet/a/**/v002/*/ma",
        "hamlet",
        "hamlet/a/**",
        "hamlet/s/sq010/sh0010/anim/**/abc",
        "hamlet/s/**/mov",
        "hamlet/s/sq010/sh0010/**/ma",
        "hamlet/s/sq001/sh0020/**",
    ]

    expandables_with_query = [
        "hamlet/s/*/**/ma?version=2",
        "hamlet/s/*/**/ma?version=v002",
        "hamlet/a/**/v002/*/ma?whatever",
        "hamlet/s/sq010/sh0010/**/ma?yes?yes",
        "hamlet/s/sq001/sh0020/**?state=w&version=>",
    ]

    for sid in expandables + expandables_with_query:
        print(" ")
        print(sid + " -->")
        results = expand(sid, do_extrapolate=False)
        # print('Expanded to -->' + str(results))
        for result in results:
            print(result.type)
            print(result.string)
            # print(sid_to_dict(result, result.type))

        pprint(results)

        print("*" * 10)



from __future__ import annotations
from typing import List, Mapping
from spil import Sid, SpilException


def from_sid_build_searches(sid: Sid | str) -> Mapping[Sid, str]:
    """
    Given a Sid, generates search sids by filling fields with * and **
    These searches can then be used for tests.

    The given sid is included in the search sids.

    :param sid:
    :return:
    """
    _sid: Sid = Sid(sid)
    if not _sid:
        raise SpilException(f'Given sid "{sid}" does not resolve to a valid Sid.')

    searches = {}
    searches[_sid] = f"Plain {_sid.type} Sid"

    for key in reversed(list(_sid.fields.keys())):

        search = _sid.get_as(key).get_with(key=key, value='**')
        searches[search] = f"Find all under {key} (**)"

        ext = _sid.get('ext') or '*'
        search = _sid.get_as(key).get_with(key=key, value='**') / ext
        searches[search] = f'Find all under {key} with "{ext}" extension'

        search = _sid.get_as(key).get_with(key=key, value='*')
        searches[search] = f"Find all {key}s"

        search = _sid.get_with(key=key, value='*')
        searches[search] = f"{_sid.type} with all {key}"

        for subkey in list(_sid.fields.keys()):
            if subkey == key:
                continue
            subsearch = search.get_with(key=subkey, value='*')
            searches[subsearch] = f"{search.type} with all {key} and {subkey}"
            for subsubkey in list(_sid.fields.keys()):
                if (subsubkey == key) or (subsubkey == subkey):
                    continue
                subsubsearch = subsearch.get_with(key=subsubkey, value='*')
                searches[subsubsearch] = f"{subsearch.type} with all {key}, {subkey} and {subsubkey}"

    return dict(reversed(sorted(searches.items())))


if __name__ == '__main__':

    from pprint import pformat
    from spil import Sid, FindInPaths, setLevel
    from spil.util.log import DEBUG, INFO, WARN, ERROR

    setLevel(ERROR)

    show_max = 10
    config = ''
    sid = Sid('hamlet/a/char/claudius/model/v001/p/ma')

    print("Starting !")

    searches = from_sid_build_searches(sid)
    for search, label in searches.items():
        print(f"search: {search} <=> \t\t\t{label}")

    print(" ")

    input("start ?")
    finder = FindInPaths(config)

    for search, label in searches.items():

        print(f"search: {search} / {label}")
        i = -1
        path_proof = ''

        for i, found in enumerate(finder.find(search)):
            bingo = ' ------> Bingo' if found == sid else ""
            if bingo or (i <= show_max):
                if config:
                    path_proof = f" ({str(found.path(config))[0:50]}...) "
                print(f"\t{found}{path_proof} {bingo}")
            if i == show_max + 1:
                print('...')

        print(f'Total found: {i+1}')
        print(' ')
        print('-' * 20)



"""
Little script running simple ad-hoc discovery tests.
"""
from spil import Sid
from spil import FindInPaths as Finder

if __name__ == "__main__":

    from spil.tests import stop
    from pprint import pprint
    from spil.util.log import setLevel, INFO

    setLevel(INFO)

    r = Sid("hamlet/s/sq010/sh0030/fx/v001/p/sword/vdb").type
    pprint(r)
    # print(Sid(path=path))
    # print(Sid("hamlet/s/sq010/sh0030/fx/v001/p/sword/vdb").path())
    # stop()

    # method chaining.
    s = (
        Sid()
        .get_with(project="hamlet", type="s")
        .get_with(query="sequence=sq010&shot=sh0010&task=anim")
        .get_as("project")
    )
    pprint(s)

    s = Sid(query="project=hamlet&type=s&sequence=sq010&shot=sh0010&task=anim")
    pprint(s)

    s = Sid("shot__sequence:hamlet/s/sq010")
    print(s.fields)

    for sid in Finder().find("hamlet/a/char/*"):
        print(f"Found: {sid}")

    # Test get_new('version')
    print("-" * 20)
    sid = Sid("hamlet/a/char/john")
    print(f"sid: {sid} - {repr(sid)}")

    print("-" * 20)
    # sid = Sid('hamlet/s/sq01/sh010/anim/*/w/ma')
    sid = Sid("hamlet/s/sq001/sh0010/anim")
    print(f"sid: {sid} - {repr(sid)}")

    print("-" * 20)
    new_scene = sid.get_with(version="*", state="w", ext="ma")
    print(f"new scene: {new_scene}")
    print(new_scene.type)

    stop()



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

# import traceback


class SpilException(Exception):
    """
    A SpilException.

    This SpilException should always be raised and handled inside this Spil package.
    """

    def __init__(self, *args, **kwargs):
        self.message = str(args[0])

    def __str__(self):
        return f"[SpilException] {self.message}"


def raiser(exception: str | Exception):
    """
    Utility function to raise errors after an "or".

    Examples:

        ... print(0 or raiser('Null value'))
        ... print(0 or raiser(ValueError('null value')))

    Inspired by https://mail.python.org/pipermail/python-ideas/2014-November/029921.html


    Args:
        exception: in case of a string, raises a SpilException

    Returns:
        raises the given exception or a SpilException
    """
    if isinstance(exception, Exception):
        raise exception
    else:
        raise SpilException(str(exception))


if __name__ == "__main__":

    try:
        try:
            raise Exception("foo")
        except Exception as ex:
            raise SpilException(ex)

    except SpilException as pe:

        print("caught : ", pe)



"""


The syntax is fieldname.EntityType.fieldname.
In this example, PlaylistVersionConnection has a field named version.
That field contains a Version entity. The field we are interested on the Version is code.
Put those together with our friend the dot and we have version.Version.code.

"""
type_mapping = {
    'project': 'Project',
    # 'asset__assettype': '',
    'asset__asset': 'Asset',
    "asset__task": 'Task',
    "asset__version": 'Version',
    'shot__sequence': 'Sequence',
    'shot__shot': 'Shot',
    "shot__task": 'Task',
    "shot__version": 'Version',
}


value_mappings = {
    'project': {
        'TestProject_TDLab': 'hamlet',
    },
    'assettype': {
        "Character": "char",
        "Prop": "prop",
        "FX": "fx",
        "Environment": "location"
    },
    'task': {
        'Layout': 'layout',
        'Animation': 'anim',
        'FX': 'fx',
        'Lighting': 'render',
        'Compositing': 'comp',
        'Art': 'art',
        'Modeling': 'model',
        'Surfacing': 'surface',
        'Rigging': 'rig',
    }
}


field_mappings = {
    'project': {
        "name": 'project'
    },
    'asset__assettype': {
        "project.Project.name": 'project',
        "sg_asset_type": "assettype",
    },
    'asset__asset': {
        "project.Project.name": 'project',
        "sg_asset_type": "assettype",
        "code": "asset"
    },
    "asset__task": {
        "project.Project.name": 'project',
        'entity.Asset.sg_asset_type': "assettype",
        "entity.Asset.code": 'asset',
        "content": "task",
    },
    "asset__version": {
        "project.Project.name": 'project',
        'entity.Asset.sg_asset_type': "assettype",
        "entity.Asset.code": 'asset',
        'sg_task.Task.content': 'task',
        "code": "version",
    },
    'shot__sequence': {
        "project.Project.name": 'project',
        "code": "sequence",
    },
    'shot__shot': {
        "project.Project.name": 'project',
        "sg_sequence.Sequence.code": 'sequence',
        "code": "shot",
    },
    "shot__task": {
        "project.Project.name": 'project',
        "entity.Shot.sg_sequence.Sequence.code": 'sequence',
        "entity.Shot.code": 'shot',
        # 'step.Step.code' : 'step' (untested)
        "content": "task",
    },
    "shot__version": {
        "project.Project.name": 'project',
        "entity.Shot.sg_sequence.Sequence.code": 'sequence',
        "entity.Shot.code": 'shot',
        # 'sg_task.Task.step.Step.code': 'step'
        'sg_task.Task.content': 'task',
        "code": "version",
    }
}

# TODO: data formatting
data_mappings = {
    'project': {
        "id": "sg_id",
        "description": 'description',
        "image": "image"
    },
    'asset__assettype': {
        "project.Project.name": 'project',
        "sg_asset_type": "assettype",
    },
    'asset__asset': {
        "description": 'description',
        "sg_note": "note",
        "id": "sg_id",
        "sg_status_list": "status",
        "shots": "shots",
        "image": "image"
    },
    "asset__task": {
        "id": "sg_id",
        "sg_status_list": "status",
        "image": "image"
    },
    "asset__version": {
        "description": 'description',
        "sg_note": "note",
        "id": "sg_id",
        "sg_status_list": "status",
        "image": "image"
    },
    'shot__sequence': {
        "id": "sg_id",
        "sg_status_list": "status",
    },
    'shot__shot': {
        "description": 'description',
        "sg_note": "note",
        "id": "sg_id",
        "sg_cut_duration": "cut_duration",
        "sg_status_list": "status",
        "assets": "assets",
        "image": "image"
    },
    "shot__task": {
        "id": "sg_id",
        "sg_status_list": "status",
        "image": "image"
    },
    "shot__version": {
        "description": 'description',
        "sg_note": "note",
        "id": "sg_id",
        "sg_status_list": "status",
        "image": "image"
    }
}



defaults_by_basetype = {
    'shot': {
        'type': 's'
    },
    'asset': {
        'type': 'a'
    }
}






"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR a PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List, Callable
from pprint import pformat

from spil import Sid
from spil.util.caching import lru_kw_cache as cache
from spil.util.log import warn, debug, error

from spil.sid.read.unfolders.extensions import execute as extensions
from spil.sid.read.unfolders.or_op import execute as or_op
from spil.sid.read.unfolders.expand import execute as expand
from spil.sid.read.unfolders.typed_narrow import execute as narrow
from spil.sid.read.unfolders.extrapolate import execute as extrapolate

# "unfolders" by execution order for existing searches
list_search_unfolders = [extensions, or_op, expand, narrow]

# FIXME: at some point we will probably need to automatically narrow the searches.
# Currently, multiple types can have identical search sids, which may lead to multiple useless runs.


def apply_unfolders(sid: str, unfolders: List[Callable]) -> List[Sid]:
    """
    Takes the given sid, executes Search Sid "unfolders" (functions)
    and returns a sorted list of unique search sids.

    "Unfolder" functions takes a list of search sids, and returns a list of search sids.

    Note: (#FIXME: this has to be refactored with explicit typing)
        The unfolder function list is: [extensions, or_op, expand, narrow]
        "extension" and "or_op" operate on lists of strings
        "expand" gets strings and returns Sid objects
        "narrow" gets Sids and returns Sids.

    Args:
        sid: a search Sid string
        unfolders: a list of callables getting a list of Sids and returning a List of Sids.

    Returns: A sorted list of unique Sid objects.

    """
    result = [sid]
    for func in unfolders:
        done = func(result)
        result = done

    return sorted(set(result))


@cache
def unfold_search(
    search_sid: str | Sid, do_uniquify: bool = False, do_extrapolate: bool = False
) -> List[Sid]:
    """
    This function transforms an untyped "search sid",
    (a Sid or string containing search signs, like "*", "**" or ",")
    into a list of typed search Sid objects.

    If do_extrapolate is True, all intermediate types are included in the result.
    Intermediate types are all the parent types of the given one (the upstream hierarchy).

    If do_uniquify is True, duplicate Sids that share the same string are removed.
    This is only useful for testing.

    Examples:
        >>> unfold_search("hamlet/s,a")
        [Sid('asset:hamlet/a'), Sid('shot:hamlet/s')]

        >>> unfold_search("hamlet/s,a/*")
        [Sid('asset__assettype:hamlet/a/*'), Sid('shot__sequence:hamlet/s/*')]

        >>> unfold_search("hamlet/s,a/*", do_extrapolate=True)
        [Sid('project:hamlet'), Sid('asset:hamlet/a'), Sid('asset__assettype:hamlet/a/*'), Sid('shot:hamlet/s'), Sid('shot__sequence:hamlet/s/*')]

        >>> unfold_search("*/*/*")
        [Sid('asset__assettype:*/a/*'), Sid('shot__sequence:*/s/*')]

        >>> unfold_search("hamlet/*/**/maya")
        [Sid('asset__file:hamlet/a/*/*/*/*/*/ma'), Sid('asset__file:hamlet/a/*/*/*/*/*/mb'), Sid('shot__file:hamlet/s/*/*/*/*/*/ma'), Sid('shot__file:hamlet/s/*/*/*/*/*/mb')]

        >>> unfold_search("hamlet/a/char/ophelia/rig")
        [Sid('asset__task:hamlet/a/char/ophelia/rig')]

    It also applies uris if present and if appliable:
        >>> unfold_search("hamlet/s/**/movie?version=>")
        [Sid('shot__movie_file:hamlet/s/*/*/*/>/*/avi'), Sid('shot__movie_file:hamlet/s/*/*/*/>/*/mov'), Sid('shot__movie_file:hamlet/s/*/*/*/>/*/mp4')]

    Args:
        search_sid: a search sid (Sid object or string)
        do_uniquify: if True remove duplicates with same string from the result. For testing purposes only.
        do_extrapolate: if True, include intermediate types in the result.

    Returns: a list of Sid objects
    """

    debug('Treating Search Sid: "{}"'.format(search_sid))

    # unfolding
    search_sids = apply_unfolders(
        str(search_sid), list_search_unfolders + ([extrapolate] if do_extrapolate else [])
    )

    # removing invalid
    for ssid in search_sids.copy():  # make this a search unfolder ?
        if not ssid:  # untyped Sid evaluates to False
            search_sids.remove(ssid)
        if ssid.string.count("?"):  # Un-appliable Query (query still present in string)
            warn(f'SearchSid "{ssid}" is typed, but has un-applied Query. Cannot be searched.')
            search_sids.remove(ssid)

    if do_uniquify:
        search_sids = uniquify_searches(search_sids)

    debug(f'Done "{search_sid}" - Unfolded {len(search_sids)} --> {pformat(search_sids)}')

    return search_sids


def uniquify_searches(search_sids: List[Sid]) -> List[Sid]:
    """
    For a list of Sids, removes duplicates having the same string representation.

    This is experimental, and should not be used in normal usage.
    Can be used during tests.

    Args:
        search_sids: a list of Sids that may contain search signs ("*", etc.)

    Returns: a new list with duplicates removed.

    """
    done = set()
    result = []
    for ssid in search_sids:
        if ssid.string not in done:
            done.add(ssid.string)
            result.append(ssid)

    return result


if __name__ == "__main__":

    from logging import FATAL
    from pprint import pprint
    from spil.util.log import setLevel, info

    info("Tests start")

    setLevel(FATAL)

    searches = {}
    searches["hamlet"] = "One project"
    searches["*"] = "All projects"
    searches["hamlet/*"] = "One project, all types "
    searches["*/s"] = "all project, one type"
    searches["*/a"] = "all project, one type"
    searches["*/s,a"] = "all project, some types"
    searches["hamlet/s,a"] = "one project, some types"
    searches["*/s,a/*"] = "all project, some types, cats / seqs"
    searches["hamlet/s,a/*"] = "one project, some types, cats / seqs"
    searches["*/*"] = "all project, all types"
    searches["*/*/*"] = "all project, all types, cats / seqs"
    searches["hamlet/a,s/*/*"] = "one project, until shots / assets"
    searches["*/*/*/*"] = "all project, until shots / assets"

    searches["hamlet/s/*/*/*"] = "one project until shot tasks"
    searches["hamlet/s/**/movie"] = "one project all shot movies"
    searches["hamlet/s/**/movie?version=>"] = "one project all shot movies, last version"
    searches["hamlet/s/**"] = ""
    searches["hamlet/*/**"] = ""
    searches["hamlet/*/**/maya"] = ""

    for test in searches.keys():

        results = unfold_search(test)
        print(test + " -->")
        pprint(results)

        results = unfold_search(test, do_extrapolate=True)
        print(test + " -->")
        pprint(results)

        print("=" * 10)



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
import json
import platform

from pathlib import Path

from spil.util.log import debug, info, error
from spil.util.singleton import Singleton

from spil.conf.global_conf import user_app_folder_name, user_conf_file_name


def get_user_config_path():
    """
    Defines the user configuration path (including the conf.json file.
    Relies on Path.home()

    Note: On windows, different python versions return different paths for Path.home().
    To be able to use this library from many Python versions,
    this was unified to be in Path.home() / Documents
    (given "Documents" exist)

    Returns a Path object.
    """
    home = Path.home()
    if home.name == "Documents" and home.parent.exists():
        home = home.parent  # fix for maya
    user_config_path = home / user_app_folder_name / user_conf_file_name

    return user_config_path


user_config_path = get_user_config_path()


class ConfigIO(Singleton):
    '''
    Writer and Reader for User config_name
    '''

    conf_path = None

    def __init__(self):

        self.conf_path = str(user_config_path)

        # Creating file
        try:
            if not user_config_path.exists():
                user_config_path.parent.mkdir(mode=0o777, parents=True, exist_ok=True)
                with open(self.conf_path, 'w') as conf_file:
                    json.dump({}, conf_file)
        except Exception as e:
            error('Problem creating Conf file : {} -> {}'.format(user_config_path, e))

    def save(self, key, value=None):

        data = self.read() or {}

        if value is None:
            data.pop(key, None)  # remove the key if None
        else:
            data[key] = value

        with open(self.conf_path, 'w') as conf_file:
            json.dump(data, conf_file)

    def read(self, key=None, default=None):

        data = {}
        try:
            with open(self.conf_path) as conf_file:
                data = json.load(conf_file)
        except Exception as e:
            error('Problem reading Conf file : {}'.format(e))

        if key is None:
            return data
        else:
            return data.get(key, default)


if __name__ == '__main__':

    from spil import logging
    logging.setLevel(logging.INFO)
    info('Path is : {}'.format(user_config_path))
    cfio = ConfigIO()
    info(cfio.read())
    cfio.save('test', 'brief test ...')
    info(cfio.read())
    cfio.save('test', None)
    cfio.save('tested', 'OK')
    info(cfio.read())




# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""


# FIXME: this is work in progress
def reload_lru_caches():
    """
    Reloads all current lru caches.


    """
    from spil.data.find_in_cache import get_sidcache
    from spil.data.data import get_cached_attribute, get_data_source
    from spil.sid.read.tools import unfold_search
    from spil.sid.core.sid_resolver import sid_to_dict, sid_to_dicts
    from spil.sid.core.fs_resolver import path_to_dict

    cached_functions = [get_cached_attribute, sid_to_dict, path_to_dict, get_data_source, unfold_search, get_sidcache, sid_to_dicts]
    for f in cached_functions:
        f.cache_clear()


if __name__ == '__main__':

    reload_lru_caches()



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List, overload, Optional
from typing_extensions import Literal

from spil.sid.sid import Sid
from spil.sid.read.util import first
from spil.sid.read.tools import unfold_search


class Finder:
    """
    Interface for Sid Finder.

    Implements common public Sid Search methods "find", "find_one", and "exists"

    The search process is as follows:
        find():
        - the search sid string is "unfolded" into a list of typed search Sids.

        do_find()
        - runs the actual search on each typed search sid
    """

    @overload
    def find(self, search_sid: str, as_sid: Literal[True]) -> Iterator[Sid]:
        ...

    @overload
    def find(self, search_sid: str, as_sid: Literal[False]) -> Iterator[str]:
        ...

    @overload
    def find(self, search_sid: Sid, as_sid: Literal[True]) -> Iterator[Sid]:
        ...

    @overload
    def find(self, search_sid: Sid, as_sid: Literal[False]) -> Iterator[str]:
        ...

    @overload
    def find(self, search_sid: str) -> Iterator[Sid]:
        ...

    @overload
    def find(self, search_sid: Sid) -> Iterator[Sid]:
        ...

    def find(
        self, search_sid: str | Sid, as_sid: Optional[bool] = True
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Yields the Sids found using the given search_sid.
        Returns a generator over Sids, if as_sid is True (default), or over Sid strings.

        Example:

            >>> from spil import FindInAll
            >>> list(FindInAll().find('hamlet/a/*'))
            [Sid('asset__assettype:hamlet/a/char'), Sid('asset__assettype:hamlet/a/location'), Sid('asset__assettype:hamlet/a/prop'), Sid('asset__assettype:hamlet/a/fx')]

        Args:
            search_sid: typed or untyped Sid or string
            as_sid: if the result should be Sid objects or strings

        Returns:
            Generator over Sids or strings
        """
        # shortcut if Sid is not a search
        sid = Sid(search_sid)
        if sid and not sid.is_search():
            generator = self.do_find([sid], as_sid=as_sid)
        else:
            search_sids = unfold_search(search_sid)
            generator = self.do_find(search_sids, as_sid=as_sid)
        yield from generator

    @overload
    def do_find(self, search_sids: List[Sid], as_sid: Literal[True]) -> Iterator[Sid]:
        ...

    @overload
    def do_find(self, search_sids: List[Sid], as_sid: Literal[False]) -> Iterator[str]:
        ...

    @overload
    def do_find(
        self, search_sids: List[Sid], as_sid: Optional[bool]
    ) -> Iterator[Sid] | Iterator[str]:
        ...

    def do_find(
        self, search_sids: List[Sid], as_sid: Optional[bool] = True
    ) -> Iterator[Sid] | Iterator[str]:
        raise NotImplementedError(
            f"[Finder.do_find] is abstract, and seams not implemented. Class: {self.__class__}"
        )

    @overload
    def find_one(self, search_sid: str | Sid, as_sid: Literal[True]) -> Sid:
        ...

    @overload
    def find_one(self, search_sid: str | Sid, as_sid: Literal[False]) -> str:
        ...

    @overload
    def find_one(self, search_sid: str | Sid, as_sid: Optional[bool]) -> Sid | str:
        ...

    def find_one(self, search_sid: str | Sid, as_sid: Optional[bool] = True) -> Sid | str:
        """
        Returns the first Sid found using the given search_sid.

        Returns a Sid, if as_sid is True (default), or a Sid strings.

        Internally calls "first" on "find".

        Example:

            >>> from spil import FindInAll
            >>> FindInAll().find_one('hamlet/a/char/ophelia')
            Sid('asset__asset:hamlet/a/char/ophelia')

        Args:
            search_sid: typed or untyped Sid or string
            as_sid: if the result should be a Sid object or string

        Returns:
            first found Sid or string

        """
        found = first(self.find(search_sid, as_sid=False))  # read is faster if as_sid is False
        if as_sid:
            return Sid(found)
        else:
            return found

    def exists(self, search_sid: str | Sid) -> bool:
        """
        Returns True if the given search_sid returns a result.
        Else False.

        Internally calls "bool" on "find_one".

        Example:

            >>> from spil import FindInAll
            >>> FindInAll().exists('hamlet/a/char/ophelia')
            True

            >>> from spil import FindInAll
            >>> FindInAll().exists('hamlet/a/char/jimmy')
            False

        Args:
            search_sid: search_sid: typed or untyped Sid or string

        Returns:
            True if search_sid returns a result, else False
        """
        return bool(self.find_one(search_sid, as_sid=False))

    def __str__(self):
        return f"[spil.{self.__class__.__name__}]"


if __name__ == "__main__":
    print(Finder())



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional, Mapping, Any
from spil import Sid

"""
As FindInAll, a configurable "WriteToAll" Writer is planned.
Note: if a FindInCache Finder is used, FindInCache.create() should also be a registered data destination.
"""


class Writer:
    """
    A "Writer" implements CRUD's Create/Update/Delete functions.

    The Writer creates Sids, with or without attributes.
    It Updates and Sets "data" / attributes for given Sids.
    It also deletes Sids.

    Writer is subclassed for the different data destinations,
    like the Filesystem or database management systems (WriteToPaths, WriteToShotgrid, etc.).

    WriteToPaths (as FindInPaths) works with Sids on the Filesystem.

    The goal of this CRUD formalisation is to propose a simple API around Sids.
    These features are still experimental.

    (Finders and Getters implement read)
    """

    def create(self, sid: Sid | str, data: Optional[dict] = None) -> bool:
        """
        Creates the given Sid, optionally with given data.

        Args:
            sid: Sid string or instance
            data: optional data dictionary

        Returns: True on success, False on failure
        """
        return False

    def update(self, sid: Sid | str, data: Mapping[str, Any]) -> bool:
        """
        Updates given data on given Sid.

        Args:
            sid: Sid string or instance
            data: data dictionary

        Returns: True on success, False on failure
        """
        return False

    def set(self, sid: Sid | str, attribute: Optional[str], value: Optional[Any], **kwargs) -> bool:
        """
        Sets given attribute with given value, and / or given kwargs, for given Sid.

        Args:
            sid: Sid string or instance
            attribute: attribute name, data key
            value: value for data
            **kwargs: key / value pairs.

        Returns: True on success, False on failure
        """
        return False

    def delete(self, sid: Sid | str) -> bool:
        """
        Deletes given Sid.

        Args:
            sid: Sid string or instance

        Returns: True on success, False on failure
        """
        return False

    def __str__(self):
        return f"[spil.{self.__class__.__name__}]"


if __name__ == "__main__":
    print(Writer())



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Mapping, List, Dict

import string
from collections import OrderedDict

from spil.util.log import debug
from spil.conf import sidtype_keytype_sep


def get_keys(template):
    return [t[1] for t in string.Formatter().parse(template) if t[1] is not None]


def extrapolate_templates(sid_templates: Mapping[str, str], to_extrapolate: List[str]) -> Mapping[str, str]:
    """
    This function receives the "sid_templates" dictionary, operates "extrapolation", and returns a modified "sid_templates" dict.
    "Extrapolation" is done for each "type" (sid type) listed in "to_extrapolate".

    "Extrapolation" means generating intermediate types (and templates) from a deeper type / longer template,
    by replacing each types "keytype" by the templates last key.

    Illustration.

        Suppose we have:
        type: shot__version, template: project/s/sequence/shot/task/version

        Extrapolation will generate:
        type: shot__task, template: project/s/sequence/shot/task
        type: shot__shot, template: project/s/sequence/shot
        type: shot__sequence, template: project/s/sequence
        type: shot__type, template: project/s
        type: shot__project, template: project

    The templates are used by the resolver.
    The goal of this function is to generate some automatically, to simplify configuration.

    Notes:
        Order of templates is important. Extrapolation starts when a "to_extrapolate" type is encountered,
        and the extrapolated types are inserted after that.
        If a type already exists in config, the generated type is skipped (not updated by extrapolation).
        If a template already exists in the config, the generated type is skipped also.

    Examples:

        >>> sid_templates = {'asset__file': '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}',\
            'asset': '{project}/{type:a}',\
            'project': '{project}'}
        >>> to_extrapolate = ['asset__file']
        >>> extrapolate_templates(sid_templates, to_extrapolate)
        OrderedDict([('asset__file', '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}'), ('asset__state', '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}'), ('asset__version', '{project}/{type:a}/{assettype}/{asset}/{task}/{version}'), ('asset__task', '{project}/{type:a}/{assettype}/{asset}/{task}'), ('asset__asset', '{project}/{type:a}/{assettype}/{asset}'), ('asset__assettype', '{project}/{type:a}/{assettype}'), ('asset', '{project}/{type:a}'), ('project', '{project}')])

        >>> sid_templates = {'asset__file': '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}',\
            'asset__state': '{project}/{type:a}/{assettype}/{asset}/{step}/{task}/{version}/{state}',\
            'asset': '{project}/{type:a}',\
            'project': '{project}'}
        >>> to_extrapolate = ['asset__state']
        >>> extrapolate_templates(sid_templates, to_extrapolate)
        OrderedDict([('asset__file', '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}'), ('asset__state', '{project}/{type:a}/{assettype}/{asset}/{step}/{task}/{version}/{state}'), ('asset__version', '{project}/{type:a}/{assettype}/{asset}/{step}/{task}/{version}'), ('asset__task', '{project}/{type:a}/{assettype}/{asset}/{step}/{task}'), ('asset__step', '{project}/{type:a}/{assettype}/{asset}/{step}'), ('asset__asset', '{project}/{type:a}/{assettype}/{asset}'), ('asset__assettype', '{project}/{type:a}/{assettype}'), ('asset', '{project}/{type:a}'), ('project', '{project}')])

    Args:
        sid_templates: A dictionary containing sid types (template names) as keys, and templates (path like strings) as values.
        to_extrapolate: List of sid types that should be extrapolated

    Returns:
        A new template dictionary containing sid_templates and the insterted extrapolated types.
    """
    new_templates = OrderedDict()

    debug("Starting template extrapolation.")

    # We loop over the existing template dictionary
    for sid_type, template in sid_templates.items():

        new_templates[sid_type] = template
        debug(f"Adding existing {sid_type} / {template}")

        # if the type is to_extrapolate
        if sid_type in to_extrapolate:
            debug(f"Extrapolating: {sid_type}")

            # getting ingredients from the sid_type and template
            keytype = sid_type.split(sidtype_keytype_sep)[-1]
            parts = template.split('/')[:-1]  # we remove the last element, that is already in the sid_type

            # walking up the parts
            for i, part in enumerate(reversed(parts)):

                # key from the template part, eg: "{type:a}" => "type"
                key = part.split(':')[0].replace('{', '').replace('}', '')

                # building the new type and template
                new_type = sid_type.replace(keytype, key)
                new_template = '/'.join(parts[:len(parts)-i])

                # we skip if template is already defined by another type
                if new_template in set(sid_templates.values()).union(set(new_templates.values())):
                    debug(f"Already defined template, skipped: {new_template}")
                    continue

                # we skip if this type was already defined
                if new_type in set(sid_templates.keys()).union(set(new_templates.keys())):
                    debug(f"Already defined type, skipped: {new_type}")
                    continue

                debug(f"Adding: {key} / {new_type} -> {new_template}")
                new_templates[new_type] = new_template

    debug("Extrapolation done.")
    return new_templates


def pattern_replacing(sid_templates: Dict[str, str], key_patterns: Dict[str, Dict[str, str]]) -> None:
    """
    Loops through the templates in sid_templates.
    Uses the key_patterns dict to replace parts of the template according to the type.

    Modifies the given template dictionary in place. Does not return anything.

    Illustration:

        For this entry in sid_templates :
            'asset__file':'{project}/{type:a}/{cat}/{name}/{task}/{version}/{state}/{ext:scenes}'

        And this entry in key_patterns:
            'asset__': {
                '{task}': r'{task:(' + '|'.join(asset_tasks) + '|\*)}',
            },

        For all types containing "asset__*", the entry {task} is replaced by r'{task:(' + '|'.join(asset_tasks) + '|\*)}',
        Meaning that the asset tasks are specified in the asset templates.

    Args:
        sid_templates: A dictionary containing sid types (template names) as keys, and templates (path like strings) as values.
        key_patterns: A dictionary, where key string is a pattern to match against a sid type, and value a replace dict.

    """

    for _keytype, template in sid_templates.copy().items():

        for m in key_patterns.keys():
            if m in _keytype:
                for find, replace in key_patterns.get(m).items():  # type: ignore
                    template = template.replace(find, replace)

        sid_templates[_keytype] = template


if __name__ == '__main__':

    from pprint import pprint
    sid_templates = {'asset__file': '{project}/{type:a}/{assettype}/{asset}/{task}/{version}/{state}/{ext:scenes}',
                     'asset': '{project}/{type:a}',
                     'project': '{project}'}
    to_extrapolate = ['asset__file']
    result = extrapolate_templates(sid_templates, to_extrapolate)
    pprint(result)

    sid_templates = {'asset__version': '{project}/{type:a}/{assettype}/{asset}/{step}/{task}/{state}/{version}', }
    to_extrapolate = ['asset__version']
    result = extrapolate_templates(sid_templates, to_extrapolate)
    pprint(result)

    import doctest
    doctest.testmod()




# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
from spil.conf.global_conf import *

# user config_name
from spil.conf.configio import ConfigIO

try:
    from spil.conf.sid_conf_load import *
    from spil.conf.data_conf_load import *
except Exception as e:
    raise Exception('Unable to import the spil_conf files (sid_conf, data_conf). \n'
                    'Please check the files compatibility with the latest SPIL version.')


# function to override config_name into user config_name
def set(key, value, save=True):
    """
    Sets a variable "key" with given "value" as a config_name variable.

    If the variable exists in the current config_name, it's value is overriden.

    If save is True (the default), the variable is saved into the user config_name, and persisted
    (unless the config_name file is wiped).

    Example :

    conf.set('prefered_project', self.ui.prod_CBB.currentText() or conf.prefered_project)
    # replaces the variable 'prefered_project' (or keeps the default one)

    See also spil.tests.conf_tests
    """
    globals()[key] = value
    if save:
        user_conf.save(key, value)


def get(key):
    """
    Gets a value from the config_name, or None if not set.
    """
    return globals().get(key)


user_conf = ConfigIO()
# if user_conf.read():
#     print(f"User config updates: {list(user_conf.read().keys())}")
globals().update(user_conf.read())

if __name__ == '__main__':

    # This will create a configuration file and folder in the user path.

    from pprint import pprint

    pprint(globals())

    project = 'test'

    # print(projects)

    # user conf
    print(project)
    set('project', 'bla')
    print(project)

    pprint(globals())



from spil_fs_conf import *  # type: ignore

# for test and demo purposes
from pathlib import Path
project_server_root_path = Path(__file__).parent / "data" / "testing" / "SPIL_PROJECTS" / "SERVER" / "PROJECTS"

path_templates = path_templates.copy()  # type: ignore

# Replace this by your own folder root, eg
# path_templates['project_root'] = r'/home/mh/Desktop/SPIL_PROJECTS/SERVER/PRJ'
path_templates['project_root'] = project_server_root_path.as_posix()





"""
This test runs and compares Finders.

Use test sids and the created test files.
"""
import random

from hamlet_scripts.example_sids import sids  # type: ignore

# from spil_plugins.sg.find_sg import FindInSG
from spil import FindInList, FindInPaths, FindInAll
from spil.tests.prep.build_searches import from_sid_build_searches
from spil.tests.utils.search_tester import check_searches_in_finder

from spil.util.log import get_logger
from spil.tests.utils.dualsearch_ab_tester import check_search_ab

log = get_logger("spil_tests", color=False)

def test_finders():

    random.seed(26)
    random.shuffle(sids)

    log.info("Starting Finder Tests")
    find_in_list = FindInList(sids)
    find_in_paths_local = FindInPaths("local")
    find_in_paths_server = FindInPaths("server")
    find_in_all = FindInAll()
    #     finder_in_sg = FindInSG()

    finders = [find_in_list, find_in_paths_local, find_in_paths_server, find_in_all]

    # sample Sids that will generate searches
    for sid in sids[:1]:

        log.info(f"Base Sid: {sid}")
        searches = from_sid_build_searches(sid)
        print("#" * 100)

        for search_sid, comment in searches.items():
            log.info(f"Running search: {search_sid}")
            search = {search_sid: comment}

            log.info("Comparing finders")
            check_search_ab(search, find_in_list, find_in_all, raise_problems=True)
            check_search_ab(search, find_in_paths_server, find_in_paths_local, raise_problems=True)

            print("-" * 50)
            for finder in finders:
                log.info(f"Running searches on {finder}")
                check_searches_in_finder(
                    search, finder, reraise=True  # , do_deep=(finder == find_in_list)
                )


if __name__ == "__main__":

    from spil.util.log import INFO
    log.setLevel(INFO)
    test_finders()



"""
This script loads test sids
and writes them into a test file.

This test file will be read in test scripts.

"""
from pathlib import Path
from spil import conf
from hamlet_scripts.generate_example_sids import sids  # type: ignore
from spil.tests.prep.save_sid_list_to_file import write_sids_to_file

sid_file = Path(conf.default_sid_conf_data_path) / "testing" / "hamlet.sids.txt"

if not sid_file.exists():
    print(f"Creating Sid test file: {sid_file}")
    write_sids_to_file(sids, sid_file)
    print(f"Done. {len(sids)} written to file")
else:
    print(f"Sid file exists, nothing written. File: {sid_file}")



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.util.log import DEBUG, get_logger
from spil import Sid, SpilException
from pprint import pformat

log = get_logger("spil_tests", color=False)
log.setLevel(DEBUG)


def check_typed_sid(s, reraise=True, replace=None):
    """
    Test protocol for the Typed Sid.

    Tests only core feature, without delegation (to Finders or Getters).
    Does not include path tests.

    Only major or unexpected problems trigger exceptions.
    Most configuration problems trigger log warnings.

    If reraise is True, exceptions are reraised, otherwise there are logged.
    replace can be a tuple that is used for string replacement on the tested Sid.

    from_search allows to pass a search string, to test the match.
    """
    s = str(s)
    if replace:
        s = s.replace(*replace)

    log.info('Testing: "{}"'.format(s))
    sid = Sid(s)
    log.info('Instanced: "{}"'.format(sid.uri))

    if not s.count("?"):  # Assert works only without Query part
        assert str(sid) == s
    assert sid == eval(repr(sid))

    try:
        if not sid.type:
            log.error('Sid "{}" not typed, skipping'.format(sid))
            return

        params = {
            "parent": sid.parent,
            "grand_parent": sid.parent.parent,
            "basetype": sid.basetype,
            "keytype": sid.keytype,
            "len": len(sid),
            "type": sid.type,
            "uri": sid.uri,
            "string": sid.string,
            "query": sid.as_query(),
            "is_search": sid.is_search(),
        }

        log.info("Params: \n" + pformat(params))
        log.info("Fields: \n" + pformat(sid.fields))

        # todo: create a Search Sid from current Sid
        """
        search_sid = sid.get_with(key=sid.keytype, value='*')
        match = True  # sid.match(search_sid)
        if not match:
            log.warning('No match "{}" <-> "{}". This is not normal'.format(sid, search_sid))
        """

        parent_key = sid.parent.keytype
        if not sid.type == "project":
            if sid.parent and sid.get_as(parent_key):
                if not sid.parent == sid.get_as(parent_key):
                    log.warning('Sid "{}" parent problem'.format(sid))
            else:
                log.warning('Sid "{}" has not parent ?'.format(sid))

        assert sid == Sid(
            sid.get("project") + "?" + sid.as_query()
        ), "Sid Query assertion pb : {}".format(Sid(sid.get("project") + "?" + sid.get_query()))

        assert sid == Sid(
            sid.get("project") + "?" + sid.as_query()
        ), "Sid Query assertion pb : {}".format(Sid(sid.get("project") + "?" + sid.as_query()))
        log.info('Passed "Query": sid == Sid(sid.get("project")) + "?" + sid.get_query() ')

        if not sid.parent:
            log.warning("Sid {} has no valid parent (got: {}).".format(sid, sid.parent))
        elif sid == sid.parent:
            log.info("Sid is a root (has no parent).")
        else:
            assert sid == sid.parent / sid.get(sid.keytype)
            log.info('Passed "parent": sid == sid.parent / sid.get(sid.keytype) ')

    except SpilException as e:
        log.error("SpilException : {} --> {}".format(e, s))
        if reraise:
            raise e

    except Exception as e:
        log.error("Exception : {} --> {}".format(e, s))
        if reraise:
            raise e


def check_typed_sids(sids, reraise=True, replace=None):
    """
    Loop over core_test_sid.
    """

    log.info("Testing if example sids match the Sid config_name")

    if not sids:
        log.warning("No sids given, nothing to test.")
        return

    for i, s in enumerate(sids):
        log.info("*" * 75)
        log.info("----------------------------------------- {}".format(i))
        check_typed_sid(s, reraise=reraise, replace=replace)


if __name__ == "__main__":

    from spil.util.log import setLevel, DEBUG, INFO

    setLevel(INFO)

    # from hamlet_scripts.example_sids import sids
    sids = []

    check_typed_sids(sids)



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Mapping, Any, Optional, List, Callable

import json

from spil.sid.read.getters.getter_finder import GetByFinder
from spil.util.log import warning, debug, error
from spil import Sid, FindInPaths
from spil.conf import default_path_config, path_data_suffix, get_data_json_path  # type: ignore


class GetFromPaths(GetByFinder):
    """
    Getter from File system.

    This is still experimental.
    """

    def __init__(self, config: Optional[str] = None):  # noqa
        self.config = config or default_path_config  # type: ignore
        self.finder = FindInPaths(config=config)

    def get_data(
        self, sid: str | Sid, attributes: Optional[List[str]] = None, sid_encode: Callable = str
    ) -> Mapping[str, Any]:
        """
        Returns data associated to the given Sid, as a key:value dictionary.
        Returns an empty dictionary if no data was found.

        Reads data dictionary from a json.
        To infer the jsons file path, the Sid path is used,
        with the suffix replaced, as per config.

        Args:
            sid:
            attributes:
            sid_encode:

        Returns: A key:value dictionary, or an empty dictionary
        """
        _sid = Sid(sid)
        sid_path = _sid.path(self.config)
        if not sid_path:
            debug(f'Given Sid "{sid}" has no path at config {self.config}. Cannot get data.')
            return {}
        data_path = get_data_json_path(sid_path)
        data: dict[str, Any] = {}
        if data_path.exists():
            try:
                with data_path.open() as f:
                    data = json.load(f) or {}
            except OSError as e:
                warning(f"Failed to open the json file. Sid: {sid}, file: {data_path}, Error: {e}")
            except json.JSONDecodeError as e:
                warning(f"Failed to decode Json. Sid: {sid}, file: {data_path}, Error: {e}")
            except Exception as e:
                warning(f"Failed to get data from json. Sid: {sid}, file: {data_path}, Error: {e}")

        encoded = sid_encode(_sid)
        if encoded:
            data["sid"] = encoded

        if attributes:
            return {key: data.get(key) for key in attributes}
        else:
            return data

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Config: "{self.config}"]'


if __name__ == "__main__":
    from pprint import pprint

    print(GetFromPaths())
    search = "hamlet/a"

    print("Calling find:")
    for result in FindInPaths().find(search):
        pprint(result)

    print("Calling get:")
    for result2 in GetFromPaths().get(search, sid_encode=lambda x: x.uri):
        pprint(result2)

    print("Calling get with lambda None:")
    for result2 in GetFromPaths().get(search, sid_encode=lambda x: None):
        pprint(result2)

    print(GetFromPaths().get_attr(search, "bill"))



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional
import os

from spil.sid.sid import Sid

from spil.util.log import info, warning, debug
from spil.sid.core import sid_resolver
from spil.sid.core.query_helper import apply_query
from spil.sid.pathops import fs_resolver


def sid_to_sid(sid: str | Sid) -> Sid:
    """
    Creates a Sid from the given Sid object or string.

    Args:
        sid: a Sid object or a string.

    Returns: a Sid object

    """
    debug(f"Starting with: {sid}")

    string = sid.uri if isinstance(sid, Sid) else str(sid)

    if string.count("?"):  # sid contains Query ending. We put it aside, and later append it back
        string, query = string.split("?", 1)
    else:
        query = ""

    # resolving
    if string.count(":"):  # a uri
        _type, string = string.split(":")
        _type, fields = sid_resolver.sid_to_dict(string, _type)
    else:  # a simple string
        _type, fields = sid_resolver.sid_to_dict(string)  # The string might be empty

    new_sid = Sid(from_factory=True)  # empty instance

    # no query to handle, we return
    if not query:
        new_sid._init(string=string, type=_type, fields=fields)
        return new_sid

    # applying the query (applying the query may update the type)
    else:
        string, _type, fields = apply_query(string, query=query, type=_type, fields=fields)
        new_sid._init(string=string, type=_type, fields=fields)
        return new_sid


def dict_to_sid(fields: dict) -> Sid | None:
    """
    Creates a Sid from a given Fields dictionary.

    Args:
        fields: a Sid fields dictionary

    Returns: Sid object

    """
    # TODO: make a fast version when the fields comes from an internal trusted and already typed call.

    debug(f"Starting with: {fields}")

    _type = sid_resolver.dict_to_type(fields)  # FIXME: terrible code
    if not _type:
        warning(f'Fields did not resolve to valid Sid type. fields: "{fields}"')
        return None

    # Now getting sid and ordered dict
    sid = sid_resolver.dict_to_sid(fields, _type)

    # TODO: this next line could be just a dict sorting ?
    type, fields = sid_resolver.sid_to_dict(sid, _type)  # check if type == _type ?

    if not fields:
        warning(f"After resolving back from Sid {sid}, Fields are empty. Check the config.")
        return None

    new_sid = Sid(from_factory=True)
    new_sid._init(string=sid, type=_type, fields=fields)
    return new_sid


def path_to_sid(path: str | os.Pathlike[str], config: Optional[str]) -> Sid | None:  # type: ignore  # (Problem with os.Pathlike)
    """
    Creates a Sid from the given path.

    Args:
        path: a path for a Sid
        config: config name for the path resolving

    Returns: Sid object

    """
    # resolving
    _type, fields = fs_resolver.path_to_dict(path, config=config)

    if not fields:
        info(f"Path [{path}] did not resolve to valid Sid fields (config_name:{config}.")
        return None

    # Now getting sid
    resolved_sid = sid_resolver.dict_to_sid(fields, _type)
    if not resolved_sid:
        info('Path "{}" did resolve to fields {}, but not back to Sid'.format(path, fields))
        return None

    new_sid = Sid(from_factory=True)
    new_sid._init(string=resolved_sid, type=_type, fields=fields)
    return new_sid


# @lru_kw_cache
def sid_factory(
    sid: Optional[str] = None,
    query: Optional[str] = None,
    fields: Optional[dict] = None,
    path: os.Pathlike[str] | str | None = None,  # type: ignore # Problem with os.Pathlike
    config: Optional[str] = None,
) -> Sid:
    """
    Sid Factory.
    Depending on input, calls a sid creation function and returns the produced Sid object.

    In case of multiple arguments, the first has priority (others are ignored). Except for path and config, that are used together.
    If no param is given, eg. Sid(), returns an empty Sid('').

    Important:
    Python always calls __init__ on objects returned by __new__.
    In this factory, we circumvent doubled __init__ calls, by having an empty __init__, and calling an explicit _init() method.
    To implement a new factory method, it is needed to explicitly call this _init(), or the DataSid object will end up malformed.

    Args:
        sid: a Sid object or string
        query: a string with "&" separated key=value pairs, eg. "?project=hamlet&type=s"
        fields: a Sid fields dictionary
        path: a path for a Sid
        config: config name for the path resolving

    Returns: Sid object

    """
    debug(f"sid_factory start: {sid} - {query} - {fields} - {path} - {config}")
    result = None
    if sid:
        result = sid_to_sid(sid)

    elif query:
        result = sid_to_sid(f"?{query}")

    elif fields:
        result = dict_to_sid(fields=fields)

    elif path:
        result = path_to_sid(path=path, config=config)

    if result is not None:
        return result
    else:
        new_sid = Sid(from_factory=True)
        new_sid._init()
        return new_sid



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""

import os
import sys

"""
This package can be imported to force the vendor versions to be used. 

The lucidity lib has some minor incompatibilities with Python 3. These were fixed in the vendor version. 

"""

vendor_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../vendor'))

if vendor_path not in sys.path:
    # print('inserting {}'.format(vendor_path))
    sys.path.insert(0, vendor_path)


if __name__ == '__main__':

    print(vendor_path)

    from pprint import pprint
    pprint(sys.path)

    import lucidity160
    print(lucidity160)

    from lucidity160 import template



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, Mapping, Any, List, overload, Optional, Callable, Dict
from typing_extensions import Literal

from spil import Sid
from spil.sid.read.util import first
from spil.sid.read.getter import Getter

from spil.util.log import debug, info, warning, error
from spil.conf import get_getter_for  # type: ignore
from spil.util.caching import lru_cache as cache
from spil.sid.read.tools import unfold_search


# @cache
def get_getter(
    sid: Sid | str, attribute: Optional[str] = None, config: Optional[str] = None
) -> Getter | None:
    """
    Calls spil.conf.get_getter_for() which is implemented in the spil_data_conf.

    Retrieves, for a given Search Sid and configuration name, the appropriate Getter and returns it.

    This typically returns a Getter depending on the Type of the Sid.
    See similar use in FindInAll and get_finder_for

    Technical note: the result is cached.
    This means that the choice of the Getter is cached, not the resulting data itself.
    The Getter is called again each time a query method (get(), get_one(), get_attr()) is called.

    Args:
        attribute:
        sid: the Search Sid for which we want to get the appropriate Getter instance
        config: an optional configuration name, to be able to have multiple configs co-existing.

    Returns:
        the Getter to use for this Search Sid

    """
    _sid = Sid(sid)
    if not str(_sid) == str(sid):
        warning(f'Sid could not be instanced, likely a configuration error. "{sid}" -> {_sid}')
    source = get_getter_for(_sid, attribute=attribute, config=config)
    if source:
        debug(f'Getting data source for "{sid}" {attribute or ""} {config or ""}): -> {source}')
        return source
    else:
        warning(f'Data Source not found for Sid "{sid}" ({_sid.type})')
        return None


class GetFromAll(Getter):  # noqa
    """
    This Getter will call other Getters, as configured, depending on the search sids type.

    The do_get() method is delegated to other Getters, and not implemented.
    """

    def __init__(self, config: Optional[str] = None):
        """
        Config is an argument that will be passed to the config, via get_getter_for(sid, config).
        Config acts like a key, to allow multiple GetFromAll configurations to co-exist.

        Args:
            config: name of a configuration
        """
        self.config = config

    def get(
        self,
        search_sid: str | Sid,
        attributes: Optional[List[str]] = None,
        sid_encode: Callable = str,
    ) -> Iterator[Mapping[str, Any]]:
        """
        See Getter.

        """
        # we start by unfolding
        search_sids: List[Sid] = unfold_search(search_sid)

        # Dictionary to map a Getter to a list of Sids it should get.
        getter_to_searches: Dict[Getter, List[Sid]] = {}

        for search_sid in search_sids:

            # Getter for the specific search
            getter = get_getter(
                search_sid, config=self.config
            )  # IDEA: allow multiple getters to cumulate data

            if getter:
                searches = getter_to_searches.get(getter) or []
                searches.append(search_sid)
                getter_to_searches[getter] = searches
            else:
                warning(f"No Getter configured for {search_sid}. Skipped from search.")

        for getter, searches in getter_to_searches.items():

            debug(f'Searching "{getter}" --> "{searches}"')
            generator = getter.do_get(searches, attributes=attributes, sid_encode=sid_encode)

            # TODO: check if doublon check is needed
            yield from generator

    def get_data(
        self, sid: str | Sid, attributes: Optional[List[str]] = None, sid_encode: Callable = str
    ) -> Mapping[str, Any]:
        """
        See Getter

        """
        # Getter for the specific sid
        source = get_getter(sid, config=self.config)
        if source:
            return source.get_data(sid, attributes=attributes, sid_encode=sid_encode)
        else:
            error(f"No Getter for get_data() found for {sid} / config: {self.config}")
            return {}

    def get_attr(self, sid: str | Sid, attribute: str) -> Any | None:
        # Getter for the specific sid and attribute
        source = get_getter(sid, attribute=attribute, config=self.config)
        if source:
            return source.get_attr(sid, attribute=attribute)
        else:
            error(
                f"No Getter for get_attr() found for {sid} and {attribute} - config: {self.config}"
            )
            return None

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Config: "{self.config}"]'


if __name__ == "__main__":
    print(GetFromAll())



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List, Set

import re

from spil import Sid
from spil.sid.read.finders.find_glob import FindByGlob
from spil.sid.core.utils import extrapolate
from spil.util.log import debug


def glob2re(pat):
    """
    Translate a shell PATTERN to a regular expression.
    There is no way to quote meta-characters.

    Borrowed from here, with many thanks.
    https://stackoverflow.com/questions/27726545/python-glob-but-against-a-list-of-strings-rather-than-the-filesystem
    where it was borrowed from fnmatch.translate

    @author: https://stackoverflow.com/users/4522780/nizam-mohamed
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i+1
        if c == '*':
            #res = res + '.*'
            res = res + '[^/]*'
        elif c == '?':
            #res = res + '.'
            res = res + '[^/]'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j+1
            if j < n and pat[j] == ']':
                j = j+1
            while j < n and pat[j] != ']':
                j = j+1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\','\\\\')
                i = j+1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return res + '\Z(?ms)'


class FindInList(FindByGlob):
    """
    List search.

    Searches for sids from a Sid string list.
    The search list is given during object instantiation.

    The search list can optionally be extrapolated.

    Implements a glob like "star_search".
    This loops through the list and accumulates the matching results.

    Note: searchlist can be a generator, but it will be exhausted after a single function call.

    Still beta.
    """
    def __init__(self, searchlist: List[str],
                 do_extrapolate: bool = False,
                 do_pre_sort: bool = False,
                 do_strip: bool = False):
        """
        Sets the search list, a list of sid strings.

        Note: searchlist could be a generator, but it would be exhausted after a single function call.
        If do_extrapolate is True, searchlist becomes a list.

        Args:
            searchlist:
            do_extrapolate:
            do_pre_sort:
            do_strip: if the returned items should be strip() (typically if coming from file input)
        """
        if do_extrapolate:
            self.searchlist = list(extrapolate(searchlist))
        else:
            self.searchlist = searchlist

        self.is_searchlist_sorted = False

        if do_pre_sort:
            self._sort_searchlist()

        self.do_strip = do_strip

    def _sort_searchlist(self):
        self.searchlist = sorted(list(set(self.searchlist)))
        debug('Pre-sorted {} sids'.format(len(self.searchlist)))
        self.is_searchlist_sorted = True

    def _get_searchlist(self, do_sort: bool = False) -> List[str]:
        if do_sort and not self.is_searchlist_sorted:
            self._sort_searchlist()
        return self.searchlist  # type: ignore
        # return (i for i in self.searchlist)  # new generator because list is used multiple times.

    def star_search(self, search_sids: List[Sid],
                    as_sid: bool = False,
                    do_sort: bool = False) -> Iterator[Sid] | Iterator[str]:
        """
        Simple star search.

        Transforms the pattern into a regex (like fnmatch.translate), but without traversing "/" (the sid separator).

        Args:
            search_sids:
            as_sid:
            do_sort:

        Returns:

        """
        done: Set[str] = set()
        done_add = done.add  # performance

        search_list = self._get_searchlist(do_sort=do_sort)

        for search_sid in search_sids:

            pattern = glob2re(str(search_sid))
            debug('[star_search] "{}"'.format(search_sid))

            for item in search_list:
                if re.match(pattern, item):
                    # debug('match : {}'.format(item))
                    if item not in done:
                        done_add(item)
                        if self.do_strip:
                            item = item.strip()
                        if as_sid:
                            yield Sid(item)
                        else:
                            yield item
                    else:
                        debug('{} was already found, skipped. '.format(item))

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- List: "{list(self.searchlist)[:5]}(...)"]'


if __name__ == "__main__":
    print(FindInList(['hamlet/s/sq010/sh0010/animation'], do_extrapolate=True))



"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2021 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.


"""
import os
import tempfile


def is_filename(path):
    """
    if the given string ends with an extension, e.g. a dot followed by 2 to 5 signs, we suppose it's a file name.

    :param path: path string
    :return:
    """
    return len(str(path).rsplit('.')) > 1 and 1 < len(str(path).rsplit('.')[-1]) < 6


def is_sequence(arg): 
    """
    Duck-typing-style sequence detection 
    
    @author Alex Martelli
    http://stackoverflow.com/questions/2937114/python-check-if-an-object-is-a-sequence 
    """
    return (not hasattr(arg, "strip") and
            hasattr(arg, "__getitem__") or
            hasattr(arg, "__iter__"))


def uniqfy(seq, reverse=False):
    """
    Returns a list with unique items, preserving order.
    
    If optional "reverse", the last duplicated item is kept. 
    (back is not optimized)
    
    http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order
    http://www.peterbe.com/plog/uniqifiers-benchmark
    
    @author Markus Jarderot
    """
    # return list(dict.fromkeys(seq))  #  This will work in Python >=3.6 and preserve order
    seen = set()
    seen_add = seen.add
    if reverse:
        result = []
        [result.insert(0, x) for x in reversed(seq) if not (x in seen or seen_add(x))]
        return result
    else:
        return [x for x in seq if not (x in seen or seen_add(x))]


def get_key(adict, value, default=None):
    """
    Gets a dict key by the value.

    If a value exists multiple times:
        If the dictionary is ordered, returns the first key.
        Otherwise key is returned "randomly".

    NB: to get a consistent result in py < 3.6, we need to use an OrderedDict as input.
    """
    try:
        index = list(adict.values()).index(value)
    except ValueError:
        return default

    result = list(adict.keys())[index]

    return result


def is_fs_case_sensitive():  # FIXME: not used, clean up.
    """
    Checks if the file system is case-sensitive.

    https://stackoverflow.com/questions/7870041/check-if-file-system-is-case-insensitive-in-python
    @author Steve Cohen, https://stackoverflow.com/users/6194443/steve-cohen

    :return: True or False
    """

    if not hasattr(is_fs_case_sensitive, 'case_sensitive'):
        with tempfile.NamedTemporaryFile(prefix='TmP') as tmp_file:
            setattr(is_fs_case_sensitive,
                    'case_sensitive',
                    not os.path.exists(tmp_file.name.lower()))

    return is_fs_case_sensitive.case_sensitive


if __name__ == '__main__':
    
    print( is_sequence('toto') )
    
    print( uniqfy(['a', 'b', 'c', 'a', 'c', 'd']) )

    print( is_filename('folder/02_260_0200-animation_export-setDressing.json') )
    print( is_filename('folder') )
    print( is_filename('test/toto.json') )

    _dict = {'name': 'Doe',
             'first_name': 'John',
             'middle_name': 'John'}
    print(get_key(_dict, 'John'))

    print('is fs case_sensitive? {}'.format(is_fs_case_sensitive()))




"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Iterator, List, Set, Optional

import os
import glob

from spil import Sid
from spil import conf
from spil.sid.pathops.pathconfig import get_path_config
from spil.sid.read.finders.find_glob import FindByGlob
from spil.util.exception import SpilException
from spil.util.log import warn, debug, error

try:
    import fileseq  # type: ignore
except ImportError:
    fileseq = None
    error("fileseq could not be imported. File sequence search will not work.")


class FindInPaths(FindByGlob):
    """

    Searches for sids in a File System.
    Mainly uses the glob search.

    Still Beta.

    """

    def __init__(self, config: Optional[str] = None):
        """
        Config is the desired config_name name, as configured in data_conf.
        If config_name is None, the default config_name (during class instantiation) is used.


        """
        self.config_name = config or conf.default_path_config  # type: ignore
        self.conf = get_path_config(self.config_name)

    def star_search(
        self, search_sids: List[Sid], as_sid: bool = False, do_sort: bool = False
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Star search main function.

        Delegates to "star_search_simple" by default,
        or to "star_search_framed" to handle file sequences, if "frame=*" is in the search

        Args:
            search_sids: list of Sids to search
            as_sid: return the result as sid object or string (the default)
            do_sort: not implemented in FindInPaths (#FIXME correct design)

        Returns:

        """
        if do_sort:
            debug("do_sort not implemented in FindInPaths")

        # depending on input, select the right generator
        # FIXME: hardcoded "frame"
        is_framed_search = any([ssid.get("frame") == "*" for ssid in search_sids])

        if is_framed_search and fileseq:
            generator = self.star_search_framed(search_sids, as_sid=as_sid)
        else:
            generator = self.star_search_simple(search_sids, as_sid=as_sid)

        yield from generator

    def star_search_simple(
        self, search_sids: List[Sid], as_sid: bool = False
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Star search without file sequence handline.

        Uses Glob.

        :param search_sids:
        :param as_sid:
        :return:
        """
        debug("Starting star_search_simple")

        # doublon detection
        searched = set()
        found_paths = set()

        for search_sid in search_sids:

            debug('Starting search:  "{}"'.format(repr(search_sid)))

            search = search_sid  # TODO: handle also strings ?

            debug("Search : {}".format(search))
            pattern = str(search.path(self.config_name))

            if not pattern:
                warn("Search sid {} did not resolve to a path. Cancelled.".format(search))
                continue

            for key, value in self.conf.search_path_mapping.items():
                pattern = pattern.replace(key, value)

            debug(f"Search pattern: {pattern}")
            if pattern in searched:
                continue
            else:
                searched.add(pattern)

            debug(f"Now searching pattern: {pattern}")
            found = glob.glob(pattern)
            debug("found")
            debug(found)
            for path in found:
                path = path.replace(os.sep, "/")
                if path in found_paths:
                    continue
                try:
                    sid = Sid(path=path, config=self.config_name)
                    debug(f"found {sid}")
                    if sid.type != search.type:
                        warn(
                            f"Found Sid and search have different types. "
                            f'Consider narrowing the pattern, or implementing "typed_search_narrowing".'
                            f"Found: {sid.uri} -- Search: {search.uri}"
                        )
                        continue
                except SpilException as e:
                    debug(f"Path did not generate sid: {path}")
                    continue
                if not sid:
                    debug(f"Path did not generate sid: {path}")
                    continue

                found_paths.add(path)
                if as_sid:
                    yield sid
                else:
                    yield str(sid)

    def star_search_framed(
        self, search_sids: List[Sid], as_sid: bool = False
    ) -> Iterator[Sid] | Iterator[str]:
        """
        Star search with file sequence handling.

        Uses Glob and fileseq.

        :param search_sids:
        :param as_sid:
        :return:
        """
        debug("Starting star_search_framed")

        searched: Set[str] = set()
        done: Set[str] = set()
        done_add = done.add

        for search_sid in search_sids:

            debug('[fs_star_search] "{}"'.format(search_sid))

            search = search_sid  # TODO: handle also strings ?

            if search.get("frame") == "*":  # FIXME: hardcoded "frame"
                search = search.get_with("frame", "@")  # for usage in fileseq

            debug("Search : " + str(search))
            pattern = str(search.path(self.config_name))

            if not pattern:
                warn("Search sid {} did not resolve to a path. Cancelled.".format(search))
                return

            if pattern in searched:
                continue
            else:
                searched.add(pattern)

            for key, value in conf.search_path_mapping.items():  # type: ignore
                pattern = pattern.replace(key, value)

            dir_pattern, file_pattern = os.path.split(pattern)
            debug(dir_pattern)
            parents = glob.glob(dir_pattern)
            file_sequences = []
            for parent in parents:
                file_search = os.path.join(parent, file_pattern).replace(os.sep, "/")
                debug("search " + file_search)
                file_sequences.extend(fileseq.findSequencesOnDisk(file_search))

            debug("found sequences : {}".format(file_sequences))
            for file_sequence in file_sequences:
                debug(file_sequence)
                path = str(file_sequence[0]).replace(
                    os.sep, "/"
                )  # we get the first file of the sequence
                try:
                    sid = Sid(path=path, config=self.config_name)
                    debug("found " + str(sid))
                except SpilException as e:
                    debug("Path did not generate sid : {}".format(path))
                    continue
                if not sid:
                    warn("Path did not generate sid : {}".format(path))
                    continue

                item = str(sid)
                if item not in done:
                    done_add(item)
                    if as_sid:
                        yield sid
                    else:
                        yield item
                else:
                    debug("{} was already found, skipped. ".format(item))

    def __str__(self):
        return f'[spil.{self.__class__.__name__} -- Config: "{self.config_name}"]'


if __name__ == "__main__":
    print(FindInPaths())



# -*- coding: utf-8 -*-
"""

This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2022 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.

"""
from spil.util.log import debug

"""
Code borrowed from the more-itertools project.
https://pypi.org/project/more-itertools/


https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/more.html#first

"""


def first(iterable, default=None):
    """Return the first item of *iterable*, or *default* if *iterable* is
    empty.

        >>> first([0, 1, 2, 3])
        0
        >>> first([], 'some default')
        'some default'

    :func:`first` is useful when you have a generator of expensive-to-retrieve
    values and want any arbitrary one. It is marginally shorter than
    ``next(iter(iterable), default)``.

    """
    try:
        return next(iter(iterable))
    except StopIteration as e:
        return default




"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
from spil.util.log import DEBUG, get_logger
from spil import Sid, SpilException
from pprint import pformat

log = get_logger("spil_tests", color=False)
log.setLevel(DEBUG)


def check_path_sid(s, configs=[], reraise=True, replace=None):
    """
    Test protocol including Path operations: Sid(path=...) and sid.path().

    """
    s = str(s)
    if replace:
        s = s.replace(*replace)

    log.info('Testing: "{}"'.format(s))
    sid = Sid(s)
    log.info('Instanced: "{}"'.format(sid.uri))

    try:
        if not sid.type:
            log.error('Sid "{}" not typed, skipping'.format(sid))
            return

        path = sid.path()
        if path:
            try:
                assert sid == Sid(path=sid.path())
                log.info("Passed : sid == Sid(path=sid.path()) ")
            except AssertionError:
                msg = "Sid path is ambiguous. {}\nsid: {}\nsid.path(): {}\nSid(path=sid.path()): {}\n".format(
                    "(Sid is a search, this may be normal)." if sid.is_search() else "",
                    sid,
                    sid.path(),
                    Sid(path=sid.path()),
                )
                log.warning(msg)
            try:
                assert sid.path() == Sid(path=sid.path()).path()
                log.info("Passed : sid.path() == Sid(path=sid.path()).path() ")
            except AssertionError:
                msg = "Sid path is ambiguous. {}\nsid: {}\nsid.path(): {}\nSid(path=sid.path()): {}\nSid(path=sid.path()).path(): {}".format(
                    "(Sid is a search, this may be normal)." if sid.is_search() else "",
                    sid,
                    sid.path(),
                    Sid(path=sid.path()),
                    Sid(path=sid.path()).path(),
                )
                log.warning(msg)
        else:
            log.error('Sid "{}" has no path.'.format(sid))

        params = {
            "path": sid.path(),
        }
        for config in configs:
            params[f"path({config})"] = sid.path(config)

        log.info("Params: \n" + pformat(params))
        log.info("Fields: \n" + pformat(sid.fields))

    except SpilException as e:
        log.error("SpilException : {} --> {}".format(e, s))
        if reraise:
            raise e

    except Exception as e:
        log.error("Exception : {} --> {}".format(e, s))
        if reraise:
            raise e


def check_path_sids(sids, configs=[], reraise=True, replace=None):
    """
    Loop over test_sid.
    """

    log.info("Testing if example sids match the Sid config_name")

    if not sids:
        log.warning("No sids given, nothing to test.")
        return

    for i, s in enumerate(sids):
        log.info("*" * 75)
        log.info("----------------------------------------- {}".format(i))
        check_path_sid(s, configs=configs, reraise=reraise, replace=replace)


if __name__ == "__main__":

    from spil.util.log import setLevel, DEBUG, INFO

    setLevel(INFO)

    # from hamlet_scripts.example_sids import sids
    sids = []

    check_path_sids(sids)



"""
This test runs tests on "path" Sid objects.
Path Sid objects implement path mapping.

Uses the example sids.
"""
import random

import spil_data_conf  # type: ignore
from hamlet_scripts.example_sids import sids  # type: ignore
from spil.tests.utils.sid_path_tester import check_path_sids


def test_path_sids():

    random.seed(26)
    random.shuffle(sids)
    tests = sids[:20]
    check_path_sids(tests, reraise=True)


if __name__ == "__main__":

    from spil.util.log import setLevel, INFO
    setLevel(INFO)
    test_path_sids()



# -*- coding: utf-8 -*-
"""
This file is part of SPIL, The Simple Pipeline Lib.

(C) copyright 2019-2023 Michael Haussmann, spil@xeo.info

SPIL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

SPIL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with SPIL.
If not, see <https://www.gnu.org/licenses/>.
"""
"""
Thank you Hugh Brown
https://gist.github.com/hughdbrown
https://gist.github.com/hughdbrown/bf5c63792d5f912a162bf012fb6b4527
Also borrowed pieces from functool.py, Copyright (C) 2006-2013 Python Software Foundation

This is a minimal but fast lru implementation.

TODO: replace for PY3 with the functools.lru_cache implementation.
Note that the cache size needs to be high enough.
"""
from functools import wraps

_max_size = 4096 * 2 * 2  # approx 40 Mo
_max_size = 4096


def lru_cache(user_function):
    cache = {}

    @wraps(user_function)
    def wrapper(*args):
        key = tuple(args)
        if key not in cache:
            # Validate we didn't exceed the max_size:
            if len(cache) >= _max_size:
                cache.popitem()
                # cache.clear()
            cache[key] = user_function(*args)
        return cache[key]

    def cache_clear():
        cache.clear()

    def cache_info():
        from pprint import pformat
        return pformat(cache)

    wrapper.cache_clear = cache_clear
    wrapper.cache_info = cache_info

    return wrapper


def lru_kw_cache(user_function):
    cache = {}
    #stats = [0, 0]  # hits,misses

    @wraps(user_function)
    def wrapper(*args, **kwargs):
        #return user_function(*args, **kwargs)
        key = tuple(args) + tuple(kwargs)
        if key not in cache:
            #stats[1] += 1  # miss
            # Validate we didn't exceed the max_size:
            if len(cache) >= _max_size:
                cache.popitem()
                # cache.clear()
            cache[key] = user_function(*args, **kwargs)
        else:
            #stats[0] += 1  # hit
            pass
        return cache[key]

    def cache_clear():
        cache.clear()
        #stats[0] = stats[1] = 0

    def cache_info():
        from pprint import pformat
        return pformat(cache)
        #return f"Hits: {stats[0]}/ Misses: {stats[1]} / Cache: {pformat(cache)}"

    wrapper.cache_clear = cache_clear
    wrapper.cache_info = cache_info
    return wrapper


def hit_cache(user_function):
    """
    A cache decorator that caches only if the wrapped function returns something.
    This is useful on data that is often queried but rarely or not at all erased.

    TODO: add TTL mechanism

    Args:
        user_function:

    Returns:

    """
    cache = {}

    @wraps(user_function)
    def wrapper(*args, **kwargs):
        key = tuple(args) + tuple(kwargs)
        if key not in cache:
            # Validate we didn't exceed the max_size:
            if len(cache) >= _max_size:
                # Delete an item in the dict:
                # print('lru_cache over size')
                cache.popitem()  # does not make sens in PY2
                # cache.clear()
            returned = user_function(*args, **kwargs)
            if returned:
                cache[key] = returned
            else:
                return returned
        return cache[key]

    def cache_clear():
        cache.clear()

    def cache_info():
        from pprint import pformat
        return pformat(cache)

    wrapper.cache_clear = cache_clear
    wrapper.cache_info = cache_info
    return wrapper



